// ========================================== //
//                                            //
//                                //
//            //
//             2017-12-23               //
//                                            //
//                HiAsm                 //
//              www.hiasm.com                 //
//                                            //
//             : Netspirit               //
//         E-mail: netspirit@meta.ua          //
//                                            //
// ========================================== //


unit BlockCipher;



interface

uses
  Windows, Kol, Share;


type

  {$ifndef KOL3XX}
  //   Delphi  FPC  NativeInt
  NativeInt = Integer;
  PNativeInt = ^NativeInt;
  NativeUInt = Cardinal;
  PNativeUInt = ^NativeUInt;
  {$endif}
  
  //      
  TBlockCryptor = class 
    protected
      FBlockSize: DWord; //     

    public
      property BlockSize: DWord read FBlockSize;
      
      constructor Create;
      
      procedure InitializeKey(Key: Pointer; KeyLen: Integer); virtual; abstract; //    
      procedure EncryptBlock(var Block); virtual; abstract; //    
      procedure DecryptBlock(var Block); virtual; abstract; //    
      procedure XORBlock(TargetBlock, KeyBlock: Pointer);
      
      //   
      procedure EncryptBufferECB(Buf: Pointer; DataLen: DWord);
      procedure DecryptBufferECB(Buf: Pointer; DataLen: DWord);
      procedure EncryptBufferCBC(Buf: Pointer; DataLen: DWord; var InitVector);
      procedure DecryptBufferCBC(Buf: Pointer; DataLen: DWord; var InitVector);
      procedure EncryptBufferCFB(Buf: Pointer; DataLen: DWord; var InitVector);
      procedure DecryptBufferCFB(Buf: Pointer; DataLen: DWord; var InitVector);
      procedure EncryptBufferOFB(Buf: Pointer; DataLen: DWord; var InitVector);
      procedure DecryptBufferOFB(Buf: Pointer; DataLen: DWord; var InitVector);
  end;
  
  TInitVector = array of Byte;
  
  // ,    
  TBlockCipher = class(TBlockCryptor)
    private
      FInitVector: TInitVector;
      FCurrentIV: TInitVector;
      
    public
      Aborted: Boolean;
      BlockMode: Integer;
      BufferSize: DWord;
      Checksum: Integer;
      ErrorCode: Integer;
      Padding: Integer;
      
      property CurrentIV: TInitVector read FCurrentIV;
      
      constructor Create;
      
      procedure ApplyInitVector(Buf: Pointer; BufSize: DWord);
      procedure ResetInitVector;
      
      //  ,        
      function CalcPaddedLength(DataLen: DWord): DWord;
      
      //        Padding ( Pad=True)
      //  Pad=True        BlockSize  
      //     
      function EncryptBufferPad(Buffer: Pointer; DataLen: DWord; Pad: Boolean): DWord;
      //        Padding ( UnPad=True)
      //     
      //   UnPad=True   Padding      -   
      function DecryptBufferPad(Buffer: Pointer; DataLen: DWord; UnPad: Boolean): DWord;
      
      //  
      function EncryptString(const SrcStr: string): string;
      function DecryptString(var SrcStr: string): Boolean;
      function EncryptStream(SrcStream, DstStream: PStream; var Count: NativeUInt): Boolean;
      function DecryptStream(SrcStream, DstStream: PStream; var Count: NativeUInt): Boolean;
  end;
  
  //     HiAsm
  ThiBlockCipher = class(TBlockCipher)
    private
      FProcessing: Boolean;
      FPassword: string;

      procedure SetBufferSize(Value: Integer);
      procedure SetInitVector(Value: string);
      procedure SetPassword(Value: string);
      
    public
      _prop_DstFileName: string;
      _prop_SrcFileName: string;
      _prop_Password: string;
      
      _data_Count: THI_Event;
      _data_Data: THI_Event;
      _data_DstFileName: THI_Event;
      _data_DstStream: THI_Event;
      _data_Password: THI_Event;
      _data_SrcFileName: THI_Event;
      _data_SrcStream: THI_Event;
      
      _event_onError: THI_Event;
      _event_onErrorPassword: THI_Event;
      _event_onResult: THI_Event;

      property _prop_BufferSize: Integer write SetBufferSize;
      property _prop_BlockMode: Integer write BlockMode;
      property _prop_Checksum: Integer write Checksum;
      property _prop_InitVector: string write SetInitVector;
      property _prop_Padding: Integer write Padding;
      
      procedure _work_doAbort(var _Data: TData; Index: Word);
      procedure _work_doBenchmark(var _Data: TData; Index: Word);
      procedure _work_doEncrypt(var _Data: TData; Index: Word);
      procedure _work_doDecrypt(var _Data: TData; Index: Word);
      procedure _work_doEncryptFile(var _Data: TData; Index: Word);  
      procedure _work_doDecryptFile(var _Data: TData; Index: Word);
      procedure _work_doEncryptStream(var _Data: TData; Index: Word);  
      procedure _work_doDecryptStream(var _Data: TData; Index: Word);
      
      procedure _work_doBlockMode(var _Data: TData; Index: Word); 
      procedure _work_doChecksum(var _Data: TData; Index: Word);
      procedure _work_doInitVector(var _Data: TData; Index: Word);
  end;
  
  
  
  
  function BSwap(const x: DWord): DWord;
  
  function ROL(const a, s: DWord): DWord;
  function ROR(const a, s: DWord): DWord;
  
  function ROLi(const a, s: DWord): DWord;
  function RORi(const a, s: DWord): DWord;
  

  function ReflectWord(W: Word): Word;
  function CalcCRC16(Buffer: Pointer; BufLen: DWord; Init: Word = $0000): Word;
  procedure FillRandom(Buffer: Pointer; Count: Integer);


const
  PADDING_NONE = 0;
  PADDING_ISO10126 = 1;
  PADDING_PKCS7 = 2;
  PADDING_BYTE = 3;
  
  MODE_ECB = 0;
  MODE_CBC = 1;
  MODE_CFB = 2;
  
  ERROR_INCORRECT_DATA = 0;
  ERROR_CANNOT_OPEN_SRC_FILE = 1;
  ERROR_CANNOT_OPEN_DST_FILE = 2;
  ERROR_STREAM_READ = 3;
  ERROR_STREAM_WRITE = 4;
  ERROR_UNKNOWN = 5;
  ERROR_CHECKSUM = 6;

  DefaultIV: array [0..3] of DWord = ($9B6C2508, $6C259B08, $25089B6C, $089B6C25);
  //DefaultBufSize = BlockSize*10240; // 160 KB

implementation

function BSwap(const x: DWord): DWord; assembler; {$ifdef FPC64} nostackframe; {$endif}
asm
  {$ifdef WIN64}
  mov eax, ecx
  {$endif}
  bswap eax
end;

function ROL(const a, s: DWord): DWord; assembler; {$ifdef FPC64} nostackframe; {$endif}
asm
  // EAX = a in 32bit
  
  {$ifdef WIN64}
  //mov eax, a
  mov eax, ecx // ECX = a in 64bit
  {$endif}
  
  //mov ecx, s
  mov ecx, edx // EDX = s // TODO: save/restore ECX?
  rol eax, cl
end;

function ROR(const a, s: DWord): DWord; assembler; {$ifdef FPC64} nostackframe; {$endif}
asm
  // EAX = a in 32bit
  
  {$ifdef WIN64}
  //mov eax, a
  mov eax, ecx // ECX = a in 64bit
  {$endif}
  
  //mov ecx, s
  mov ecx, edx // EDX = s // TODO: save/restore ECX?
  ror eax, cl
end;


{$ifdef VER120}

function ROLi(const a, s: DWord): DWord; assembler;
asm
  mov ecx, s
  rol eax, cl
end;

function RORi(const a, s: DWord): DWord; assembler;
asm
  mov ecx, s
  ror eax, cl
end;

{$else}

function ROLi(const a, s: DWord): DWord; inline;
begin
  Result := (a shl s) or (a shr (32 - s));
end;

function RORi(const a, s: DWord): DWord; inline;
begin
  Result := (a shr s) or (a shl (32 - s));
end;

{$endif}


function ReflectWord(W: Word): Word;
var
  I: Byte;
begin
  Result := 0;
  for I := 0 to 15 do
  begin
    Result := Result shl 1;
    if W and 1 <> 0 then Result := Result or 1;
    W := W shr 1;
  end;
end;


// Poly: $8005
// RefIn=True
// RefOut=True
function CalcCRC16(Buffer: Pointer; BufLen: DWord; Init: Word = $0000): Word;
const
  Poly = $8005;
  InvPoly = $A001; // = ReflectWord($8005);
var
  B, J: Byte;
begin
  Result := Init;
  while BufLen > 0 do
  begin
    B := Byte(Buffer^);

    // 1) RefIn=False
    {Result := Result xor (B shl 8);
    J := 0;
    while J < 8 do
    begin
      if (Result and $8000) <> 0 
      then
        Result := (Result shl 1) xor Poly
      else
        Result := (Result shl 1);
      Inc(J);
    end;}
    
    // 2) RefIn=True
    Result := Result xor B;
    J := 0;
    while J < 8 do
    begin
      if (Result and 1) <> 0 
      then
        Result := (Result shr 1) xor InvPoly
      else
        Result := (Result shr 1);
      Inc(J);
    end;
    
    Inc(NativeUInt(Buffer));
    Dec(BufLen);
  end;
  {
     RefOut=True   1  
      ,  RefOut=False - 
  }
  //Result := ReflectWord(Result); 
end;

procedure FillRandom(Buffer: Pointer; Count: Integer);
begin
  while Count > 0 do
  begin
    Byte(Buffer^) := Byte(Random(255));
    Inc(NativeUInt(Buffer));
    Dec(Count);
  end;
end;

// =============== TBlockCryptor =============== //

constructor TBlockCryptor.Create;
begin
  inherited;
  InitializeKey(nil, 0);  
end;

procedure TBlockCryptor.XORBlock(TargetBlock, KeyBlock: Pointer);
var
  I: Integer;
begin
  for I := 0 to (BlockSize div 4) - 1 do
  begin
    Integer(TargetBlock^) := Integer(TargetBlock^) xor Integer(KeyBlock^);
    Inc(NativeUInt(TargetBlock), SizeOf(Integer));
    Inc(NativeUInt(KeyBlock), SizeOf(Integer));
  end;
end;

procedure TBlockCryptor.EncryptBufferECB(Buf: Pointer; DataLen: DWord);
var
  BS: DWord;
begin
  BS := BlockSize;
  while DataLen >= BS do
  begin
    EncryptBlock(Buf^);
    Dec(DataLen, BS);
    Inc(NativeUInt(Buf), BS);
  end;
end;

procedure TBlockCryptor.DecryptBufferECB(Buf: Pointer; DataLen: DWord);
var
  BS: DWord;
begin
  BS := BlockSize;
  while DataLen >= BS do
  begin
    DecryptBlock(Buf^);
    Dec(DataLen, BS);
    Inc(NativeUInt(Buf), BS);
  end;
end;

procedure TBlockCryptor.EncryptBufferCBC(Buf: Pointer; DataLen: DWord; var InitVector);
var
  BS: DWord;
begin
  BS := BlockSize;
  while DataLen >= BS do
  begin
      XORBlock(Buf, @InitVector);
    EncryptBlock(Buf^);
      Move(Buf^, InitVector, BS);
    Dec(DataLen, BS);
    Inc(NativeUInt(Buf), BS);
  end; 
end;

procedure TBlockCryptor.DecryptBufferCBC(Buf: Pointer; DataLen: DWord; var InitVector);
var
  BS: DWord;
  IV: TInitVector;
begin
  BS := BlockSize;
  SetLength(IV, BS);
  while DataLen >= BS do
  begin
      Move(InitVector, IV[0], BS);
      Move(Buf^, InitVector, BS);
    DecryptBlock(Buf^);
      XORBlock(Buf, @IV[0]);
    Dec(DataLen, BS);
    Inc(NativeUInt(Buf), BS);
  end; 
end;

procedure TBlockCryptor.EncryptBufferCFB(Buf: Pointer; DataLen: DWord; var InitVector);
var
  BS: DWord;
begin
  BS := BlockSize;
  while DataLen >= BS do
  begin
      EncryptBlock(InitVector);
      XORBlock(@InitVector, Buf);
      Move(InitVector, Buf^, BS);
    Dec(DataLen, BS);
    Inc(NativeUInt(Buf), BS);
  end; 
end;

procedure TBlockCryptor.DecryptBufferCFB(Buf: Pointer; DataLen: DWord; var InitVector);
var
  BS: DWord;
  C: TInitVector;
begin
  BS := BlockSize;
  SetLength(C, BS);
  while DataLen >= BS do
  begin
      Move(Buf^, C[0], BS);
      Move(InitVector, Buf^, BS);
      EncryptBlock(Buf^);
      XORBlock(Buf, @C[0]);
      Move(C[0], InitVector, BS);
    Dec(DataLen, BS);
    Inc(NativeUInt(Buf), BS);
  end; 
end;

procedure TBlockCryptor.EncryptBufferOFB(Buf: Pointer; DataLen: DWord; var InitVector);
var
  BS: DWord;
begin
  BS := BlockSize;
  while DataLen >= BS do
  begin
      EncryptBlock(InitVector);
      XORBlock(Buf, @InitVector);
    Dec(DataLen, BS);
    Inc(NativeUInt(Buf), BS);
  end; 
end;

procedure TBlockCryptor.DecryptBufferOFB(Buf: Pointer; DataLen: DWord; var InitVector);
begin
  EncryptBufferOFB(Buf, DataLen, InitVector);
end;

// =========================================== //



// =============== TBlockCipher =============== //

constructor TBlockCipher.Create;
begin
  inherited;
  BufferSize := BlockSize*10240;
  ApplyInitVector(nil, 0);
  Padding := PADDING_ISO10126;
end;

procedure TBlockCipher.ApplyInitVector(Buf: Pointer; BufSize: DWord);
begin
  if BufSize > BlockSize then BufSize := BlockSize;
  
  if DWord(Length(FInitVector)) <> BlockSize then
    SetLength(FInitVector, BlockSize);
  
  if (BufSize = 0) or (Buf = nil) then
    Move(DefaultIV[0], FInitVector[0], BlockSize)
  else
  begin
    if BufSize < BlockSize then FillChar(FInitVector[BufSize], (BlockSize - BufSize), #0);
    Move(Buf^, FInitVector[0], BufSize);
  end;
end;

procedure TBlockCipher.ResetInitVector;
begin
  if Length(FCurrentIV) = 0 then
    FCurrentIV := Copy(FInitVector)
  else
    Move(FInitVector[0], FCurrentIV[0], BlockSize);
end;

function TBlockCipher.CalcPaddedLength(DataLen: DWord): DWord;
begin
  if Padding <> PADDING_NONE then
    Result := DataLen + (BlockSize - DataLen mod BlockSize)
  else
    Result := DataLen - (DataLen mod BlockSize); //   
end;

function TBlockCipher.EncryptBufferPad(Buffer: Pointer; DataLen: DWord; Pad: Boolean): DWord;
var
  C: Byte;
begin
  
  if Pad and (Padding <> PADDING_NONE) then
  begin
    C := BlockSize - (DataLen mod BlockSize); //      
    case Padding of
      PADDING_ISO10126:
        begin
          FillRandom(Pointer(NativeUInt(Buffer) + DataLen), C);
          PByte(NativeUInt(Buffer) + DataLen + C - 1)^ := C;
        end;
      PADDING_PKCS7:
        begin
          FillChar(Pointer(NativeUInt(Buffer) + DataLen)^, C, C);
        end;
      PADDING_BYTE:
        begin
          FillChar(Pointer(NativeUInt(Buffer) + DataLen)^, C, 0);
          PByte(NativeUInt(Buffer) + DataLen)^ := $80;
        end;
    end;
    Result := DataLen + C; //   
  end
  else
    Result := DataLen;
  
  case BlockMode of
    MODE_CBC: EncryptBufferCBC(Buffer, Result, FCurrentIV[0]);
    MODE_CFB: EncryptBufferCFB(Buffer, Result, FCurrentIV[0]);
    else      EncryptBufferECB(Buffer, Result);
  end;
  
end;

function TBlockCipher.DecryptBufferPad(Buffer: Pointer; DataLen: DWord; UnPad: Boolean): DWord;
var
  C: Byte;
begin
  
  DataLen := DataLen - (DataLen mod BlockSize); //   
  
  case BlockMode of
    MODE_CBC: DecryptBufferCBC(Buffer, DataLen, FCurrentIV[0]);
    MODE_CFB: DecryptBufferCFB(Buffer, DataLen, FCurrentIV[0]);
    else      DecryptBufferECB(Buffer, DataLen);
  end;
  Result := DataLen;
  
  if UnPad and (Padding <> PADDING_NONE) and (DataLen > 0) {and
     (DataLen mod BlockSize = 0)}
  then
  begin
    C := 0;
    case Padding of
      PADDING_ISO10126, PADDING_PKCS7:
        begin
          C := PByte(NativeUInt(Buffer) + DataLen - 1)^;
          if C > BlockSize then C := 0; //   
          //  PADDING_PKCS7   ,    C   C
        end;
      {PADDING_PKCS7:
        begin
          C := PByte(NativeUInt(Buffer) + DataLen - 1)^;
          if C > BlockSize then C := 0;
        end;}
      PADDING_BYTE:
        begin
          //   $80  
          Inc(NativeUInt(Buffer), DataLen);
          C := 1;
          while C <= BlockSize do
          begin
            if PByte(NativeUInt(Buffer) - C)^ = $80 then Break;
            //   ,    0
            Inc(C);
          end;
          if C > BlockSize then C := 0; //  $80  
        end;
    end;
    Dec(Result, C); //   
  end;
  
end;

function TBlockCipher.EncryptString(const SrcStr: string): string;
var
  C, L: Integer;
begin
  ResetInitVector;
  L := BinaryLength(SrcStr);
  
  if Checksum <> 0 then
  begin
    C := CalcPaddedLength(L + 2); //  PADDING_NONE      ( 0),  
    SetLength(Result, BinaryStrSize(C)); //        
    PadBinaryBuf(Pointer(Result), C);
    if C > 0 then // C = 0  PADDING_NONE  (  + 2)  BlockSize
    begin
      if L > (C - 2) then L := C - 2; //  PADDING_NONE
      
      if L > 0 then 
      begin
        Move(SrcStr[1], Result[{$ifdef UNICODE}2{$else}3{$endif}], L);
        C := Min((BlockSize*2 - 2), L); // .  .      2 
        C := CalcCRC16(@Result[{$ifdef UNICODE}2{$else}3{$endif}], C);
      end
      else
        C := 0; //   0-   = $0000
      PWord(@Result[1])^ := C;
      Inc(L, 2);
    end;
  end
  else
  begin
    C := CalcPaddedLength(L);
    SetLength(Result, BinaryStrSize(C));
    PadBinaryBuf(Pointer(Result), C);
    if C > 0 then // C = 0  PADDING_NONE  (  + 2)  BlockSize
    begin
      if L > C then L := C; //  PADDING_NONE
      Move(SrcStr[1], Result[1], L);
    end;
  end;
  
  EncryptBufferPad(Pointer(Result), L, True);
end;

function TBlockCipher.DecryptString(var SrcStr: string): Boolean;
var
  L, C: Integer;
begin
  Result := False;
  ErrorCode := ERROR_UNKNOWN;
  ResetInitVector;
  
  UniqueString(SrcStr); //   
  L := BinaryLength(SrcStr);
  
  if (Padding <> PADDING_NONE) and ((L = 0) or (DWord(L) mod BlockSize <> 0)) then
  begin
    ErrorCode := ERROR_INCORRECT_DATA;
    Exit;
  end;
  
  C := DecryptBufferPad(Pointer(SrcStr), L, True);

  //     
  if (Checksum <> 0) and (C > 0) then //   -  CRC  2- 
  begin
    L := C;
    
    C := Min(BlockSize*2, L) - 2; // .  .      2 
    if (C < 0) or
       (PWord(@SrcStr[1])^ <> CalcCRC16(@SrcStr[{$ifdef UNICODE}2{$else}3{$endif}], C))
    then
    begin
      ErrorCode := ERROR_CHECKSUM;
      Exit;
    end;
    //SrcStr := Copy(SrcStr, {$ifdef UNICODE}2{$else}3{$endif}, (L-2) div SizeOf(Char)); //  . 
    SrcStr := BufToBinaryStr(@SrcStr[{$ifdef UNICODE}2{$else}3{$endif}], L-2); //  . 
  end
  else
  begin
  
    {if (Padding <> PADDING_NONE) and (C = L) then // Padding ,     -   
    begin
      ErrorCode := ERROR_INCORRECT_DATA;
      Exit;
    end;}
    
    //L := C;
    SetLength(SrcStr, {L} BinaryStrSize(C)); //     
    PadBinaryBuf(Pointer(Result), C);
  end;

  Result := True;
end;

function TBlockCipher.EncryptStream(SrcStream, DstStream: PStream; var Count: NativeUInt): Boolean;
var
  Buf: array of Byte;
  I, R, NeedR, W, NeedW, TotalW: NativeUInt;
  C: Byte;
  Last: Boolean;
label
  finish;
begin
  Result := False;
  TotalW := 0;
  Aborted := False;
  ErrorCode := ERROR_UNKNOWN;
  
  ResetInitVector;
  
  SetLength(Buf, BufferSize+BlockSize); //    1   (    )

  I := 0;
  
  Last := False; //     
  NeedW := 0;
  
  repeat
    NeedR := Count - I; //   
    if NeedR > NativeUInt(BufferSize) then NeedR := NativeUInt(BufferSize);
    //NeedR := Min(Count - I, BufferSize);
    
    R := 0;
    if (I = 0) and (Checksum <> 0) then //   
    begin
      if NeedR > 0 then
      begin
        if (NeedR + 2 > NativeUInt(BufferSize)) then NeedR := NativeUInt(BufferSize) - 2; //     2   . 
        R := SrcStream.Read(Buf[2], NeedR);
        
        if R = 0 then //    
        begin
          ErrorCode := ERROR_STREAM_READ;
          goto finish;
        end;
      end;
      {else
        R := 0;}
      
      NeedW := R + 2;
      
      C := Min((BlockSize * 2) - 2, R);
      PWord(@Buf[0])^ := CalcCRC16(@Buf[2], C);
    end
    else
    begin
      if NeedR > 0 then
      begin
        R := SrcStream.Read(Buf[0], NeedR);
        NeedW := R;
        
        if R = 0 then
        begin
          ErrorCode := ERROR_STREAM_READ;
          goto finish;
        end;
      end;
      {else
        R := 0;}
    end;
   
    if Aborted then Break;
    
    Inc(I, R);
    
    Last := (I = Count) or (R < NeedR);
    
    //  ( ,   )
    NeedW := EncryptBufferPad(@Buf[0], NeedW, Last);
    
    if Aborted then Break;
    
    //    
    W := DstStream.Write(Buf[0], NeedW);
    
    Inc(TotalW, W);
    
    if W < NeedW then //   
    begin
      ErrorCode := ERROR_STREAM_WRITE;
      goto finish;
    end;
    
  until Aborted or Last {(R < NeedR) or (I = Count)}; // ,      ,  
  
  Result := True;
  
finish:
  Count := TotalW;
end;

function TBlockCipher.DecryptStream(SrcStream, DstStream: PStream; var Count: NativeUInt): Boolean;
var
  Buf: array of Byte;
  I, R, NeedR, W, NeedW, TotalW: NativeUInt;
  C: Byte;
  Last: Boolean;
label
  finish;
begin
  Result := False;
  Aborted := False;
  TotalW := 0;
  
  if (Padding <> PADDING_NONE) and ((Count = 0) or (Count mod NativeUInt(BlockSize) <> 0)) then
  begin
    ErrorCode := ERROR_INCORRECT_DATA;
    goto finish;
  end;
  
  ErrorCode := ERROR_UNKNOWN;
  
  ResetInitVector;
  
  SetLength(Buf, BufferSize);

  I := 0;
  Last := False;
  
  while not (Aborted or Last) do
  begin
    NeedR := Count - I; //   
    if NeedR > Cardinal(BufferSize) then NeedR := BufferSize;
    //NeedR := Min(Count - I, BufferSize);
    
    if NeedR > 0 then
    begin
      R := SrcStream.Read(Buf[0], NeedR);
      
      if Aborted then Break;
      
      if R = 0 then //    
      begin
        ErrorCode := ERROR_STREAM_READ;
        goto finish;
      end;

      Inc(I, R);
    end
    else
      R := 0;
    
    Last := (I = Count) or (R < NeedR);
    
    //   (  ,   )
    NeedW := DecryptBufferPad(@Buf[0], R, Last);
    
    if Aborted or (NeedW = 0) then Break; // NeedW = 0  PADDING_NONE  R < BlockSize
    
    {if Last and (Padding <> PADDING_NONE) and (NeedW = R) then // Padding ,     -   
    begin
      ErrorCode := ERROR_INCORRECT_DATA;
      goto finish;
    end;}
    

    if (TotalW = 0) and (Checksum <> 0) then //      
    begin
      C := BlockSize * 2;
      if NeedW < C then C := Byte(NeedW);
      Dec(C, 2);
      //C := Min((BlockSize * 2), NeedW) - 2;
      
      if PWord(@Buf[0])^ <> CalcCRC16(@Buf[2], C) then
      begin
        ErrorCode := ERROR_CHECKSUM;
        SrcStream.Seek(-I, spCurrent); //     
        goto finish;
      end;
      
      Dec(NeedW, 2);
      W := DstStream.Write(Buf[2], NeedW); //       . 
    end
    else
      W := DstStream.Write(Buf[0], NeedW); //    
    
    Inc(TotalW, W);
  
    if W < NeedW then //  
    begin
      ErrorCode := ERROR_STREAM_WRITE;
      goto finish;
    end;
    
  end;
  
  Result := True;
  
  
finish:
  Count := TotalW;
end;

// =========================================== //



// =============== ThiBlockCipher =============== //


procedure ThiBlockCipher.SetBufferSize(Value: Integer);
begin
  BufferSize := (DWord(Value) div BlockSize)*BlockSize; //      BlockSize
  if BufferSize < BlockSize*10240 then BufferSize := BlockSize*10240;
end;

procedure ThiBlockCipher.SetInitVector(Value: string);
begin
  ApplyInitVector(Pointer(Value), BinaryLength(Value));
end;

procedure ThiBlockCipher.SetPassword(Value: string);
begin
  if Value = FPassword then Exit;
  FPassword := Value;
  InitializeKey(Pointer(FPassword), BinaryLength(FPassword));
end;

procedure ThiBlockCipher._work_doAbort;
begin
  FProcessing := False;
  Aborted := True;
end;


procedure ThiBlockCipher._work_doBenchmark;
var
  I: Integer;
  Buf: array of Byte;
begin
  I := ToInteger(_Data);
  if I < 1 then I := 1;
  
  Aborted := False;
  
  SetPassword(ReadString(_Data, _data_Password, _prop_Password));

  SetLength(Buf, BufferSize);
  while (I > 0) and (not Aborted) do
  begin
    EncryptBufferECB(@Buf[0], BufferSize);
    Dec(I);
  end;
end;

procedure ThiBlockCipher._work_doEncrypt;
var 
  S, Pass: string;
begin
  if FProcessing then Exit;
  
  S := ReadString(_Data, _data_Data, '');
  
  Pass := ReadString(_Data, _data_Password, _prop_Password);
  if Pass = '' then
  begin
    _hi_CreateEvent(_Data, @_event_onErrorPassword);
    Exit;
  end;
  
  FProcessing := True;
  
  SetPassword(Pass);
  
  _hi_CreateEvent(_Data, @_event_onResult, EncryptString(S));
  
  FProcessing := False;
end;

procedure ThiBlockCipher._work_doDecrypt;
var
  S, Pass: string;
begin
  if FProcessing then Exit;
  
  S := ReadString(_Data, _data_Data, '');
  
  Pass := ReadString(_Data, _data_Password, _prop_Password);
  if Length(Pass) = 0 then
  begin
    _hi_CreateEvent(_Data, @_event_onErrorPassword);
    Exit;
  end;
  
  FProcessing := True;
  
  SetPassword(Pass);
  
  if DecryptString(S) then
    _hi_CreateEvent(_Data, @_event_onResult, S)
  else
  begin
    if ErrorCode = ERROR_CHECKSUM then
      _hi_CreateEvent(_Data, @_event_onErrorPassword)
    else
      _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  end;
  
  FProcessing := False;
end;

procedure ThiBlockCipher._work_doEncryptFile;
var 
  Src, Dst: PStream;
  C: NativeUInt;
  Pass, Fn: string;
  Del: Boolean;
label
  finish;
begin
  if FProcessing then Exit;
  
  Del := False;
  
  Pass := ReadString(_Data, _data_Password, _prop_Password);
  if Length(Pass) = 0 then
  begin
    _hi_CreateEvent(_Data, @_event_onErrorPassword);
    Exit;
  end; 
  
  SetPassword(Pass);
  
  Dst := nil;
  
  Src := NewReadFileStream(ReadString(_Data, _data_SrcFileName, _prop_SrcFileName));
  if Src.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_SRC_FILE);
    goto finish;
  end;
  
  Fn := ReadString(_Data, _data_DstFileName, _prop_DstFileName);
  Dst := NewFileStream(Fn, ofOpenWrite or ofCreateAlways or ofShareDenyWrite);
  if Dst.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_DST_FILE);
    goto finish;
  end;
  
  FProcessing := True;
  
  C := Src.Size;
  if EncryptStream(Src, Dst, C) then
  begin
    if not Aborted then
      _hi_CreateEvent(_Data, @_event_onResult, NativeUInt(C))
    else
      Del := True;
  end
  else
  begin
    Del := True;
    _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  end;
  
  FProcessing := False;
  
  
finish:
  Src.Free;
  Dst.Free;
  
  if Del then DeleteFile(PChar(Fn));
end;


procedure ThiBlockCipher._work_doDecryptFile;
var 
  Src, Dst: PStream;
  C: NativeUInt;
  Pass, Fn: string;
  Del: Boolean;
label
  finish;
begin
  if FProcessing then Exit;
  Del := False;
  
  Pass := ReadString(_Data, _data_Password, _prop_Password);
  if Length(Pass) = 0 then
  begin
    _hi_CreateEvent(_Data, @_event_onErrorPassword);
    Exit;
  end; 
  
  SetPassword(Pass);
  
  Dst := nil;
  
  Src := NewReadFileStream(ReadString(_Data, _data_SrcFileName, _prop_SrcFileName));
  if Src.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_SRC_FILE);
    goto finish;
  end;
  
  Fn := ReadString(_Data, _data_DstFileName, _prop_DstFileName);
  Dst := NewFileStream(Fn, ofOpenWrite or ofCreateAlways or ofShareDenyWrite);
  if Dst.Handle = INVALID_HANDLE_VALUE then
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_DST_FILE);
    goto finish;
  end;
  
  FProcessing := True;
  
  C := Src.Size;
  if DecryptStream(Src, Dst, C) then
  begin
    if not Aborted then
      _hi_CreateEvent(_Data, @_event_onResult, NativeUInt(C))
    else
      Del := True;
  end
  else
  begin
    Del := True;
    if ErrorCode = ERROR_CHECKSUM then
      _hi_CreateEvent(_Data, @_event_onErrorPassword)
    else
      _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  end;
  
  FProcessing := False;
  
  
finish:
  Src.Free;
  Dst.Free;
  
  if Del then DeleteFile(PChar(Fn));
end;


procedure ThiBlockCipher._work_doEncryptStream;
var 
  Src, Dst: PStream;
  C: NativeUInt;
  Pass: string;
begin
  if FProcessing then Exit;
  
  Pass := ReadString(_Data, _data_Password, _prop_Password);
  if Length(Pass) = 0 then
  begin
    _hi_CreateEvent(_Data, @_event_onErrorPassword);
    Exit;
  end; 
  
  SetPassword(Pass);
  
  Src := ReadStream(_Data, _data_SrcStream);
  Dst := ReadStream(_Data, _data_DstStream);
  C := ReadInteger(_Data, _data_Count, 0);
  
  if Src = nil then 
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_SRC_FILE);
    Exit;
  end;
  
  if Dst = nil then 
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_DST_FILE);
    Exit;
  end;

  SetPassword(Pass);
  
  FProcessing := True;
  
  if EncryptStream(Src, Dst, C) then
  begin
    if not Aborted then _hi_CreateEvent(_Data, @_event_onResult, NativeUInt(C));
  end
  else
    _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  
  FProcessing := False;
end;

procedure ThiBlockCipher._work_doDecryptStream;
var 
  Src, Dst: PStream;
  C: NativeUInt;
  Pass: string;
begin
  if FProcessing then Exit;
  
  Pass := ReadString(_Data, _data_Password, _prop_Password);
  if Length(Pass) = 0 then
  begin
    _hi_CreateEvent(_Data, @_event_onErrorPassword);
    Exit;
  end; 
  
  SetPassword(Pass);
  
  Src := ReadStream(_Data, _data_SrcStream);
  Dst := ReadStream(_Data, _data_DstStream);
  C := ReadInteger(_Data, _data_Count, 0);
  
  if Src = nil then 
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_SRC_FILE);
    Exit;
  end;
  
  if Dst = nil then 
  begin
    _hi_CreateEvent(_Data, @_event_onError, ERROR_CANNOT_OPEN_DST_FILE);
    Exit;
  end;

  SetPassword(Pass);
  
  FProcessing := True;
  
  if DecryptStream(Src, Dst, C) then
  begin
    if not Aborted then _hi_CreateEvent(_Data, @_event_onResult, NativeUInt(C));
  end
  else
  begin
    if ErrorCode = ERROR_CHECKSUM then
      _hi_CreateEvent(_Data, @_event_onErrorPassword)
    else
      _hi_CreateEvent(_Data, @_event_onError, ErrorCode);
  end;
  
  FProcessing := False;
end;


procedure ThiBlockCipher._work_doBlockMode;
begin
  _prop_BlockMode := ToInteger(_Data);
end;

procedure ThiBlockCipher._work_doChecksum;
begin
  _prop_Checksum := ToInteger(_Data);
end;

procedure ThiBlockCipher._work_doInitVector;
begin
  _prop_InitVector := Share.ToString(_Data);
end;

// =========================================== //



initialization
  Randomize;

end.