// ========================================================================================= //
//                                                                                           //
//                         AES (Rijndael)                     //
//                                     2017-12-21                                      //
//                                                                                           //
//                                       HiAsm                                         //
//                                     www.hiasm.com                                         //
//                                                                                           //
//        : Netspirit                                                                   //
//        E-mail: netspirit@meta.ua                                                          //
//                                                                                           //
//                                                       //
//                                                                                           //
//        1)    " ".                      //
//              .                                                //
//        2)          ,                     //
//                ,                  //
//            .                                                                     //
//        3)                             //
//           (  )   .                            //
//                                //
//                .                                //
//        4)                          //
//                                   //
//                  .                   //
//        5)        (, )         //
//                .                                //
//                                                                                           //
//                        ,                              //
//                                                                  //
//                                                                                           //
//                                           //
//                                                                                           //
// ========================================================================================= //

// *********************************** //
//                                     //
//        RIJNDAEL algorithm           //
//            Copyright                //
//   Joan Daemen and Vincent Rijmen    //
//                                     //
//   Used optimized variant from:      //
//          Vincent Rijmen             //
//        Antoon Bosselaers            //
//          Paulo Barreto              //
//                                     //
// *********************************** //


unit hiAESCipher;



interface





uses 
  BlockCipher;


const
  BLOCK_SIZE   = 16;
  WORDS_IN_BLOCK = BLOCK_SIZE div SizeOf(LongWord); // Nb  
  
  MAX_KEY_LENGTH = 32;
  MAX_ROUNDS = 14;
  MAX_SUBKEY_COUNT = WORDS_IN_BLOCK*(MAX_ROUNDS + 1); // = 60

type
  
  TAESBlock = array[0..WORDS_IN_BLOCK-1] of LongWord;
  
  ThiAESCipher = class(ThiBlockCipher)
    private
      Rounds: Integer; // Nr  
      //   : WORDS_IN_BLOCK * (Rounds + 1)
      EKey, DKey: array[0..MAX_SUBKEY_COUNT-1] of LongWord;
      
    public
      constructor Create;
      procedure InitializeKey(Key: Pointer; KeyLen: Integer); override;
      procedure EncryptBlock(var Block); override;
      procedure DecryptBlock(var Block); override;
  end;





implementation

var

  {
    SBox[x] = S [x].[01, 01, 01, 01];
    Te0[x] = S [x].[02, 01, 01, 03];
    Te1[x] = S [x].[03, 02, 01, 01];
    Te2[x] = S [x].[01, 03, 02, 01];
    Te3[x] = S [x].[01, 01, 03, 02];

    SBoxInv[x] = Si[x].[01, 01, 01, 01];
    Td0[x] = Si[x].[0e, 09, 0d, 0b];
    Td1[x] = Si[x].[0b, 0e, 09, 0d];
    Td2[x] = Si[x].[0d, 0b, 0e, 09];
    Td3[x] = Si[x].[09, 0d, 0b, 0e];
  }
  
  SBox: array [0..255] of LongWord;
  Te0: array [0..255] of LongWord;
  Te1: array [0..255] of LongWord;
  Te2: array [0..255] of LongWord;
  Te3: array [0..255] of LongWord;
  
  SBoxInv: array [0..255] of LongWord;
  Td0: array [0..255] of LongWord;
  Td1: array [0..255] of LongWord;
  Td2: array [0..255] of LongWord;
  Td3: array [0..255] of LongWord;

const
  { for 128-bit blocks, Rijndael never uses more than 10 Rcon values }
  Rcon: array[0..9] of LongWord = 
  (
    $01000000, $02000000, $04000000, $08000000, $10000000,
    $20000000, $40000000, $80000000, $1B000000, $36000000
  );


procedure GenerateTables;
var
  I, T: Integer;
  P, Q: Byte;
  
  PowTab: array [0..255] of Byte;
  LogTab: array [0..255] of Byte;
  
  function FFMult(A, B: Byte): LongWord;
  begin
    if {(A <> 0) and} (B <> 0) then 
      Result := PowTab[(LogTab[A] + LogTab[B]) mod 255]
    else
      Result := 0;
  end;  
begin
  P := 1;
  for I := 0 to 255 do
  begin
    PowTab[I] := P;
    LogTab[P] := I;
    
    if (P and $80) <> 0 then
      P := P xor (P shl 1) xor $01b
    else
      P := P xor (P shl 1);
  end;
  
  LogTab[1] := 0;
  
  {P := 1;
  for I := 0 to 9 do
  begin
    Rcon[I] := P;
    
    if (P and $80) <> 0 then
      P := (P shl 1) xor $1b
    else
      P := P shl 1;
  end;}
  
  for I := 0 to 255 do
  begin
    if I > 0 then
      P := PowTab[255 - LogTab[I]]
    else
      P := 0;
    
    Q := P;
    
    Q := (Q shr 7) or (Q shl 1);
    P := P xor Q;
    Q := (Q shr 7) or (Q shl 1);
    P := P xor Q;
    Q := (Q shr 7) or (Q shl 1);
    P := P xor Q;
    Q := (Q shr 7) or (Q shl 1);
    P := P xor Q xor $63;
    
    SBox[I] := P or (P shl 8) or (P shl 16) or (P shl 24);
    SBoxInv[P] := I or (I shl 8) or (I shl 16) or (I shl 24);
  end;

  for I := 0 to 255 do
  begin
    P := SBox[I];

    T := (FFMult(2, P) shl 24) or (P shl 16) or (P shl 8) or FFMult(3, P);
    
    Te0[I] := T;
    Te1[I] := ROR(T, 8);
    Te2[I] := ROR(T, 16);
    Te3[I] := ROR(T, 24);

    
    P := SBoxInv[I];
    
    T := (FFMult(14, P) shl 24) or (FFMult(9, P) shl 16) or (FFMult(13, P) shl 8) or FFMult(11, P);
    
    Td0[I] := T;
    Td1[I] := ROR(T, 8);
    Td2[I] := ROR(T, 16);
    Td3[I] := ROR(T, 24);
  end;
end;

// Rounds count by Key size and Block size:
{
|     Key size: |  16  24  32 |
|---------------|-------------|
|Block size: 16 |  10  12  14 |
|            24 |  12  12  14 |
|            32 |  14  14  14 |
|---------------|-------------|
}

{function GetRounds(const KeyLength: Integer): Integer;
begin
  case (KeyLength) of
    16:
      if BLOCK_SIZE = 16 then Result := 10
      else
        if BLOCK_SIZE = 24 then Result := 12
        else Result := 14;
    24:
      if BLOCK_SIZE <> 32 then Result := 12
      else Result := 14;
    else // 32 bytes = 256 bits
      Result := 14;
  end;
end;}




constructor ThiAESCipher.Create;
begin
  FBlockSize := BLOCK_SIZE;
  inherited;
end;

procedure ThiAESCipher.InitializeKey(Key: Pointer; KeyLen: Integer);
var
  i, t, j: LongWord;
  Nk: Integer;
begin
  Rounds := 0;
  
  if KeyLen > MAX_KEY_LENGTH then KeyLen := MAX_KEY_LENGTH;

  Nk := (KeyLen + 3) div 4;
  
  if Nk <= 4 then Nk := 4
  else
    if Nk <= 6 then Nk := 6
    else
      Nk := 8;
  
  FillChar(EKey[0], Length(EKey)*SizeOf(LongWord), 0);
  
  if KeyLen > 0 then
  begin
    Move(Key^, EKey[0], KeyLen);
    for i := 0 to Nk - 1 do
    begin
      EKey[i] := BSwap(EKey[i]);
    end;
  end;
  
  //Rounds := GetRounds(Nk*4);
  
  case Nk of
    4:
      begin
        Rounds := 10;
        for i := 0 to 9 do
        //for i := 0 to (4 * Rounds div Nk) - 1 do
        begin
          t := EKey[4 * i + 3];
          
          EKey[4 * i + 4] := EKey[4 * i] xor
            (SBox[(t shr 16) and $ff] and $ff000000) xor
            (SBox[(t shr  8) and $ff] and $00ff0000) xor
            (SBox[(t       ) and $ff] and $0000ff00) xor
            (SBox[(t shr 24)        ] and $000000ff) xor
             Rcon[i];
            //(Rcon[i] shl 24);
            

          EKey[4 * i + 5] := EKey[4 * i + 1] xor EKey[4 * i + 4];
          EKey[4 * i + 6] := EKey[4 * i + 2] xor EKey[4 * i + 5];
          EKey[4 * i + 7] := EKey[4 * i + 3] xor EKey[4 * i + 6];
        end;
      end;
    6:
      begin
        Rounds := 12;
        for i := 0 to 7 do
        begin
          t := EKey[6 * i + 5];
          t := 
            (SBox[(t shr 16) and $ff] and $ff000000) xor
            (SBox[(t shr  8) and $ff] and $00ff0000) xor
            (SBox[(t       ) and $ff] and $0000ff00) xor
            (SBox[(t shr 24)        ] and $000000ff) xor
             Rcon[i];
          
          t := t xor EKey[6 * i];
          EKey[6 * i + 6] := t;
          t := t xor EKey[6 * i + 1];
          EKey[6 * i + 7] := t;
          t := t xor EKey[6 * i + 2];
          EKey[6 * i + 8] := t;
          t := t xor EKey[6 * i + 3];
          EKey[6 * i + 9] := t;
          t := t xor EKey[6 * i + 4];
          EKey[6 * i + 10] := t;
          t := t xor EKey[6 * i + 5];
          EKey[6 * i + 11] := t;
        end;
      end;
    8:
      begin
        Rounds := 14;
        for i := 0 to 6 do
        begin
          t := EKey[8 * i + 7];
          t := 
            (SBox[(t shr 16) and $ff] and $ff000000) xor
            (SBox[(t shr  8) and $ff] and $00ff0000) xor
            (SBox[(t       ) and $ff] and $0000ff00) xor
            (SBox[(t shr 24)        ] and $000000ff) xor
            Rcon[i];
          
          t := t xor EKey[8 * i];
          EKey[8 * i + 8] := t;
          t := t xor EKey[8 * i + 1];
          EKey[8 * i + 9] := t;
          t := t xor EKey[8 * i + 2];
          EKey[8 * i + 10] := t;
          t := t xor EKey[8 * i + 3];
          EKey[8 * i + 11] := t;
          
          t := 
            (SBox[(t shr 24)        ] and $ff000000) xor
            (SBox[(t shr 16) and $ff] and $00ff0000) xor
            (SBox[(t shr  8) and $ff] and $0000ff00) xor
            (SBox[(t       ) and $ff] and $000000ff);
          
          t := t xor EKey[8 * i + 4];
          EKey[8 * i + 12] := t;
          t := t xor EKey[8 * i + 5];
          EKey[8 * i + 13] := t;
          t := t xor EKey[8 * i + 6];
          EKey[8 * i + 14] := t;
          t := t xor EKey[8 * i + 7];
          EKey[8 * i + 15] := t;
        end;
      end;
  end;
  
  { invert the order of the round keys: }
  i := 0;
  j := 4 * Rounds;
  
  while i <= j do
  begin
    DKey[i    ] := EKey[j    ]; DKey[j    ] := EKey[i    ];
    DKey[i + 1] := EKey[j + 1]; DKey[j + 1] := EKey[i + 1];
    DKey[i + 2] := EKey[j + 2]; DKey[j + 2] := EKey[i + 2];
    DKey[i + 3] := EKey[j + 3]; DKey[j + 3] := EKey[i + 3];
    Inc(i, 4);
    Dec(j, 4);
  end;
  
  { apply the inverse MixColumn transform to all round keys but the first and the last: }

  for i := 1 to Rounds - 1 do
  begin
    t := DKey[4*i    ];
    DKey[4*i    ] :=
      Td0[SBox[(t shr 24)        ] and $ff] xor
      Td1[SBox[(t shr 16) and $ff] and $ff] xor
      Td2[SBox[(t shr  8) and $ff] and $ff] xor
      Td3[SBox[(t       ) and $ff] and $ff];
    
    t := DKey[4*i + 1];
    DKey[4*i + 1] :=
      Td0[SBox[(t shr 24)        ] and $ff] xor
      Td1[SBox[(t shr 16) and $ff] and $ff] xor
      Td2[SBox[(t shr  8) and $ff] and $ff] xor
      Td3[SBox[(t       ) and $ff] and $ff];

    t := DKey[4*i + 2];
    DKey[4*i + 2] :=
      Td0[SBox[(t shr 24)        ] and $ff] xor
      Td1[SBox[(t shr 16) and $ff] and $ff] xor
      Td2[SBox[(t shr  8) and $ff] and $ff] xor
      Td3[SBox[(t       ) and $ff] and $ff];

    t := DKey[4*i + 3];
    DKey[4*i + 3] :=
      Td0[SBox[(t shr 24)        ] and $ff] xor
      Td1[SBox[(t shr 16) and $ff] and $ff] xor
      Td2[SBox[(t shr  8) and $ff] and $ff] xor
      Td3[SBox[(t       ) and $ff] and $ff];
  end;
end;


procedure ThiAESCipher.EncryptBlock(var Block);
var
  B: TAESBlock absolute Block;
  s0, s1, s2, s3, t0, t1, t2, t3: LongWord;
  i: Integer;
begin
  s0 := BSwap(B[0]) xor EKey[0];
  s1 := BSwap(B[1]) xor EKey[1];
  s2 := BSwap(B[2]) xor EKey[2];
  s3 := BSwap(B[3]) xor EKey[3];
  
  {
    (Rounds - 1) full rounds:
  }

  for i := 1 to Rounds - 1 do
  begin
    t0 :=
      Te0[(s0 shr 24)        ] xor
      Te1[(s1 shr 16) and $ff] xor
      Te2[(s2 shr  8) and $ff] xor
      Te3[(s3       ) and $ff] xor (EKey[4*i    ]);
    t1 :=
      Te0[(s1 shr 24)        ] xor
      Te1[(s2 shr 16) and $ff] xor
      Te2[(s3 shr  8) and $ff] xor
      Te3[(s0       ) and $ff] xor (EKey[4*i + 1]);
    t2 :=
      Te0[(s2 shr 24)        ] xor
      Te1[(s3 shr 16) and $ff] xor
      Te2[(s0 shr  8) and $ff] xor
      Te3[(s1       ) and $ff] xor (EKey[4*i + 2]);
    t3 :=
      Te0[(s3 shr 24)        ] xor
      Te1[(s0 shr 16) and $ff] xor
      Te2[(s1 shr  8) and $ff] xor
      Te3[(s2       ) and $ff] xor (EKey[4*i + 3]);
    
    s0 := t0;
    s1 := t1;
    s2 := t2;
    s3 := t3;
  end;
  
  {
    apply last round and
    map cipher state to byte array block:
  }
  
  i := 4*Rounds;
  t0 :=
    (SBox[(s0 shr 24)        ] and $ff000000) xor
    (SBox[(s1 shr 16) and $ff] and $00ff0000) xor
    (SBox[(s2 shr  8) and $ff] and $0000ff00) xor
    (SBox[(s3       ) and $ff] and $000000ff) xor
    EKey[i   ];
  t1 :=
    (SBox[(s1 shr 24)        ] and $ff000000) xor
    (SBox[(s2 shr 16) and $ff] and $00ff0000) xor
    (SBox[(s3 shr  8) and $ff] and $0000ff00) xor
    (SBox[(s0       ) and $ff] and $000000ff) xor
    EKey[i + 1];
  t2 :=
    (SBox[(s2 shr 24)        ] and $ff000000) xor
    (SBox[(s3 shr 16) and $ff] and $00ff0000) xor
    (SBox[(s0 shr  8) and $ff] and $0000ff00) xor
    (SBox[(s1       ) and $ff] and $000000ff) xor
    EKey[i + 2];
  t3 :=
    (SBox[(s3 shr 24)        ] and $ff000000) xor
    (SBox[(s0 shr 16) and $ff] and $00ff0000) xor
    (SBox[(s1 shr  8) and $ff] and $0000ff00) xor
    (SBox[(s2       ) and $ff] and $000000ff) xor
    EKey[i + 3];
  
  B[0] := BSwap(t0);
  B[1] := BSwap(t1);
  B[2] := BSwap(t2);
  B[3] := BSwap(t3);
end;

procedure ThiAESCipher.DecryptBlock(var Block);
var
  B: TAESBlock absolute Block;
  s0, s1, s2, s3, t0, t1, t2, t3: LongWord;
  i: Integer;
begin
  s0 := BSwap(B[0]) xor DKey[0];
  s1 := BSwap(B[1]) xor DKey[1];
  s2 := BSwap(B[2]) xor DKey[2];
  s3 := BSwap(B[3]) xor DKey[3];
  
  {
    (Rounds - 1) full rounds:
  }
  
  for i := 1 to Rounds - 1 do
  begin
    t0 :=
      Td0[(s0 shr 24)        ] xor
      Td1[(s3 shr 16) and $ff] xor
      Td2[(s2 shr  8) and $ff] xor
      Td3[(s1       ) and $ff] xor (DKey[4*i    ]);
    t1 :=
      Td0[(s1 shr 24)        ] xor
      Td1[(s0 shr 16) and $ff] xor
      Td2[(s3 shr  8) and $ff] xor
      Td3[(s2       ) and $ff] xor (DKey[4*i + 1]);
    t2 :=
      Td0[(s2 shr 24)        ] xor
      Td1[(s1 shr 16) and $ff] xor
      Td2[(s0 shr  8) and $ff] xor
      Td3[(s3       ) and $ff] xor (DKey[4*i + 2]);
    t3 :=
      Td0[(s3 shr 24)        ] xor
      Td1[(s2 shr 16) and $ff] xor
      Td2[(s1 shr  8) and $ff] xor
      Td3[(s0       ) and $ff] xor (DKey[4*i + 3]);
    
    s0 := t0;
    s1 := t1;
    s2 := t2;
    s3 := t3;
  end;
  
  {
    complete the last round and
    map cipher state to byte array block:
  }
  
  i := 4*Rounds;
  t0 :=
    (SBoxInv[(s0 shr 24)        ] and $ff000000) xor
    (SBoxInv[(s3 shr 16) and $ff] and $00ff0000) xor
    (SBoxInv[(s2 shr  8) and $ff] and $0000ff00) xor
    (SBoxInv[(s1       ) and $ff] and $000000ff);
  t1 :=
    (SBoxInv[(s1 shr 24)        ] and $ff000000) xor
    (SBoxInv[(s0 shr 16) and $ff] and $00ff0000) xor
    (SBoxInv[(s3 shr  8) and $ff] and $0000ff00) xor
    (SBoxInv[(s2       ) and $ff] and $000000ff);
  t2 :=
    (SBoxInv[(s2 shr 24)        ] and $ff000000) xor
    (SBoxInv[(s1 shr 16) and $ff] and $00ff0000) xor
    (SBoxInv[(s0 shr  8) and $ff] and $0000ff00) xor
    (SBoxInv[(s3       ) and $ff] and $000000ff);
  t3 :=
    (SBoxInv[(s3 shr 24)        ] and $ff000000) xor
    (SBoxInv[(s2 shr 16) and $ff] and $00ff0000) xor
    (SBoxInv[(s1 shr  8) and $ff] and $0000ff00) xor
    (SBoxInv[(s0       ) and $ff] and $000000ff);

  
  B[0] := BSwap(t0 xor (DKey[i    ]));
  B[1] := BSwap(t1 xor (DKey[i + 1]));
  B[2] := BSwap(t2 xor (DKey[i + 2]));
  B[3] := BSwap(t3 xor (DKey[i + 3]));
end;


initialization
  GenerateTables;
  
  
end.