// ========================================================================================= //
//                                                                                           //
//                           Serpent                          //
//                                    2016-10-29                                       //
//                                                                                           //
//                                      HiAsm                                          //
//                                    www.hiasm.com                                          //
//                                                                                           //
//        : Netspirit                                                                   //
//        E-mail: netspirit@meta.ua                                                          //
//                                                                                           //
//                                                       //
//                                                                                           //
//        1)    " ".                      //
//              .                                                //
//        2)          ,                     //
//                ,                  //
//            .                                                                     //
//        3)                             //
//           (  )   .                            //
//                                //
//                .                                //
//        4)                          //
//                                   //
//                  .                   //
//        5)        (, )         //
//                .                                //
//                                                                                           //
//                        ,                              //
//                                                                  //
//                                                                                           //
//                                           //
//                                                                                           //
// ========================================================================================= //

            // ************************************************************** //
            //                       Serpent Algorithm                        //
            //   Copyright (C) 1998 Ross Anderson, Eli Biham, Lars Knudsen.   //
            // ************************************************************** //



unit hiSerpentCipher;



interface




uses 
  BlockCipher;

const
  ROUNDS = 32; // Do not change!
  BLOCK_SIZE = 16;
  SUBKEY_COUNT = 4 * (ROUNDS + 1); // 132
  
type

  TSerpentBlock = array [0..(BLOCK_SIZE div 4)-1] of LongWord;
  
  ThiSerpentCipher = class(ThiBlockCipher)
    private
      SubKeys: array [0..SUBKEY_COUNT - 1] of LongWord;
      
    public
      constructor Create;
      procedure InitializeKey(Key: Pointer; KeyLen: Integer); override;
      procedure EncryptBlock(var Block); override;
      procedure DecryptBlock(var Block); override;
  end;





implementation


const

  MAX_KEY_LENGTH = 32;
  
  // PHI: Constant used in the key schedule
  PHI = $9e3779b9;


// S0:   3  8 15  1 10  6  5 11 14 13  4  2  7  0  9 12 //
// depth = 5,7,4,2, Total gates=18 //
procedure RND00(var _in, _out);
var
  t02, t03, t05, t06, 
  t07, t08, t09, t11, 
  t12, t13, t14, t15, 
  t17, t01 : LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[1] xor Bi[2];
  t02 := Bi[0] or Bi[3];
  t03 := Bi[0] xor Bi[1];
  Bo[3] := t02 xor t01;

  t05 := Bi[2] or Bo[3];
  t06 := Bi[0] xor Bi[3];
  t07 := Bi[1] or Bi[2];
  t08 := Bi[3] and t05;
  t09 := t03   and t07;
  Bo[2] := t09 xor t08;

  t11 := t09   and Bo[2];
  t12 := Bi[2] xor Bi[3];
  t13 := t07   xor t11;
  t14 := Bi[1] and t06;
  t15 := t06   xor t13;
  Bo[0] :=     not t15;
  
  t17 := Bo[0] xor t14;
  Bo[1] := t12 xor t17;
end;

// InvS0:  13  3 11  0 10  6  5 12  1 14  4  7 15  9  8  2 //
// depth = 8,4,3,6, Total gates=19 //
procedure InvRND00(var _in, _out);
var
  t02, t03, t04, t05, 
  t06, t08, t09, t10, 
  t12, t13, t14, t15, 
  t17, t18, t01 : LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[2] xor Bi[3];
  t02 := Bi[0] or Bi[1];
  t03 := Bi[1] or Bi[2];
  t04 := Bi[2] and t01;
  t05 := t02   xor t01;
  t06 := Bi[0] or t04;
  Bo[2] :=     not t05;

  t08 := Bi[1] xor Bi[3];
  t09 := t03   and t08;
  t10 := Bi[3] or Bo[2];
  Bo[1] := t09 xor t06;

  t12 := Bi[0] or t05;
  t13 := Bo[1] xor t12;
  t14 := t03   xor t10;
  t15 := Bi[0] xor Bi[2];
  Bo[3] := t14 xor t13;
  
  t17 := t05   and t13;
  t18 := t14   or t17;
  Bo[0] := t15 xor t18;
end;



// S1:  15 12  2  7  9  0  5 10  1 11 14  8  6 13  3  4 //
// depth = 10,7,3,5, Total gates=18 //
procedure RND01(var _in, _out);
var
  t02, t03, t04, t05, 
  t06, t07, t08, t10, 
  t11, t12, t13, t16, 
  t17, t01 : LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] or Bi[3];
  t02 := Bi[2] xor Bi[3];
  t03 :=       not Bi[1];
  t04 := Bi[0] xor Bi[2];
  t05 := Bi[0] or t03;
  t06 := Bi[3] and t04;
  t07 := t01   and t02;
  t08 := Bi[1] or t06;
  Bo[2] := t02 xor t05;
  t10 := t07   xor t08;
  t11 := t01   xor t10;
  t12 := Bo[2] xor t11;
  t13 := Bi[1] and Bi[3];
  Bo[3] :=     not t10;
  Bo[1] := t13 xor t12;
  t16 := t10   or Bo[1];
  t17 := t05   and t16;
  Bo[0] := Bi[2] xor t17;
end;


// InvS1:   5  8  2 14 15  6 12  3 11  4  7  9  1 13 10  0 //
// depth = 7,4,5,3, Total gates=18 //
procedure InvRND01(var _in, _out);
var
  t02, t03, t04, t05, 
  t06, t07, t08, t09, 
  t10, t11, t14, t15, 
  t17, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] xor Bi[1];
  t02 := Bi[1] or Bi[3];
  t03 := Bi[0] and Bi[2];
  t04 := Bi[2] xor t02;
  t05 := Bi[0] or t04;
  t06 := t01   and t05;
  t07 := Bi[3] or t03;
  t08 := Bi[1] xor t06;
  t09 := t07   xor t06;
  t10 := t04   or t03;
  t11 := Bi[3] and t08;
  Bo[2] :=     not t09;
  Bo[1] := t10 xor t11;
  t14 := Bi[0] or Bo[2];
  t15 := t06   xor Bo[1];
  Bo[3] := t01 xor t04;
  t17 := Bi[2] xor t15;
  Bo[0] := t14 xor t17;
end;


// S2:   8  6  7  9  3 12 10 15 13  1 14  4  0 11  5  2 //
// depth = 3,8,11,7, Total gates=16 //
procedure RND02(var _in, _out);
var
  t02, t03, t05, t06, 
  t07, t08, t09, t10, 
  t12, t13, t14, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] or Bi[2];
  t02 := Bi[0] xor Bi[1];
  t03 := Bi[3] xor t01;
  Bo[0] := t02 xor t03;
  t05 := Bi[2] xor Bo[0];
  t06 := Bi[1] xor t05;
  t07 := Bi[1] or t05;
  t08 := t01   and t06;
  t09 := t03   xor t07;
  t10 := t02   or t09;
  Bo[1] := t10 xor t08;
  t12 := Bi[0] or Bi[3];
  t13 := t09   xor Bo[1];
  t14 := Bi[1] xor t13;
  Bo[3] :=     not t09;
  Bo[2] := t12 xor t14;
end;


// InvS2:  12  9 15  4 11 14  1  2  0  3  6 13  5  8 10  7 //
// depth = 3,6,8,3, Total gates=18 //
procedure InvRND02(var _in, _out);
var
  t02, t03, t04, t06, 
  t07, t08, t09, t10, 
  t11, t12, t15, t16, 
  t17, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] xor Bi[3];
  t02 := Bi[2] xor Bi[3];
  t03 := Bi[0] and Bi[2];
  t04 := Bi[1] or t02;
  Bo[0] := t01 xor t04;
  t06 := Bi[0] or Bi[2];
  t07 := Bi[3] or Bo[0];
  t08 :=       not Bi[3];
  t09 := Bi[1] and t06;
  t10 := t08   or t03;
  t11 := Bi[1] and t07;
  t12 := t06   and t02;
  Bo[3] := t09 xor t10;
  Bo[1] := t12 xor t11;
  t15 := Bi[2] and Bo[3];
  t16 := Bo[0] xor Bo[1];
  t17 := t10   xor t15;
  Bo[2] := t16 xor t17;
end;

// S3:   0 15 11  8 12  9  6  3 13  1  2  4 10  7  5 14 //
// depth = 8,3,5,5, Total gates=18 //
procedure RND03(var _in, _out);
var
  t02, t03, t04, t05, 
  t06, t07, t08, t09, 
  t10, t11, t13, t14, 
  t15, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] xor Bi[2];
  t02 := Bi[0] or Bi[3];
  t03 := Bi[0] and Bi[3];
  t04 := t01   and t02;
  t05 := Bi[1] or t03;
  t06 := Bi[0] and Bi[1];
  t07 := Bi[3] xor t04;
  t08 := Bi[2] or t06;
  t09 := Bi[1] xor t07;
  t10 := Bi[3] and t05;
  t11 := t02   xor t10;
  Bo[3] := t08 xor t09;
  t13 := Bi[3] or Bo[3];
  t14 := Bi[0] or t07;
  t15 := Bi[1] and t13;
  Bo[2] := t08 xor t11;
  Bo[0] := t14 xor t15;
  Bo[1] := t05 xor t04;
end;

// InvS3:   0  9 10  7 11 14  6 13  3  5 12  2  4  8 15  1 //
// depth = 3,6,4,4, Total gates=17 //
procedure InvRND03(var _in, _out);
var
  t02, t03, t04, t05, 
  t06, t07, t09, t11, 
  t12, t13, t14, t16, 
  t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[2]   or Bi[3];
  t02 := Bi[0]   or Bi[3];
  t03 := Bi[2]   xor t02;
  t04 := Bi[1]   xor t02;
  t05 := Bi[0]   xor Bi[3];
  t06 := t04     and t03;
  t07 := Bi[1]   and t01;
  Bo[2] := t05   xor t06;
  t09 := Bi[0]   xor t03;
  Bo[0] := t07   xor t03;
  t11 := Bo[0]   or t05;
  t12 := t09     and t11;
  t13 := Bi[0]   and Bo[2];
  t14 := t01     xor t05;
  Bo[1] := Bi[1] xor t12;
  t16 := Bi[1]   or t13;
  Bo[3] := t14   xor t16;
end;

// S4:   1 15  8  3 12  0 11  6  2  5  4 10  9 14  7 13 //
// depth = 6,7,5,3, Total gates=19 //
procedure RND04(var _in, _out);
var
  t02, t03, t04, t05, 
  t06, t08, t09, t10, 
  t11, t12, t13, t14, 
  t15, t16, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] or Bi[1];
  t02 := Bi[1] or Bi[2];
  t03 := Bi[0] xor t02;
  t04 := Bi[1] xor Bi[3];
  t05 := Bi[3] or t03;
  t06 := Bi[3] and t01;
  Bo[3] := t03 xor t06;
  t08 := Bo[3] and t04;
  t09 := t04   and t05;
  t10 := Bi[2] xor t06;
  t11 := Bi[1] and Bi[2];
  t12 := t04   xor t08;
  t13 := t11   or t03;
  t14 := t10   xor t09;
  t15 := Bi[0] and t05;
  t16 := t11   or t12;
  Bo[2] := t13 xor t08;
  Bo[1] := t15 xor t16;
  Bo[0] :=     not t14;
end;

// InvS4:   5  0  8  3 10  9  7 14  2 12 11  6  4 15 13  1 //
// depth = 6,4,7,3, Total gates=17 //
procedure InvRND04(var _in, _out);
var
  t02, t03, t04, t05, t06, 
  t07, t09, t10, t11, t12, 
  t13, t15, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[1] or Bi[3];
  t02 := Bi[2] or Bi[3];
  t03 := Bi[0] and t01;
  t04 := Bi[1] xor t02;
  t05 := Bi[2] xor Bi[3];
  t06 :=       not t03;
  t07 := Bi[0] and t04;
  Bo[1] := t05 xor t07;
  t09 := Bo[1] or t06;
  t10 := Bi[0] xor t07;
  t11 := t01   xor t09;
  t12 := Bi[3] xor t04;
  t13 := Bi[2] or t10;
  Bo[3] := t03 xor t12;
  t15 := Bi[0] xor t04;
  Bo[2] := t11 xor t13;
  Bo[0] := t15 xor t09;
end;

// S5:  15  5  2 11  4 10  9 12  0  3 14  8 13  6  7  1 //
// depth = 4,6,8,6, Total gates=17 //
procedure RND05(var _in, _out);
var
  t02, t03, t04, t05, 
  t07, t08, t09, t10, 
  t11, t12, t13, t14, 
  t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[1] xor Bi[3];
  t02 := Bi[1] or Bi[3];
  t03 := Bi[0] and t01;
  t04 := Bi[2] xor t02;
  t05 := t03   xor t04;
  Bo[0] :=     not t05;
  t07 := Bi[0] xor t01;
  t08 := Bi[3] or Bo[0];
  t09 := Bi[1] or t05;
  t10 := Bi[3] xor t08;
  t11 := Bi[1] or t07;
  t12 := t03   or Bo[0];
  t13 := t07   or t10;
  t14 := t01   xor t11;
  Bo[2] := t09 xor t13;
  Bo[1] := t07 xor t08;
  Bo[3] := t12 xor t14;
end;

// InvS5:   8 15  2  9  4  1 13 14 11  6  5  3  7 12 10  0 //
// depth = 4,6,9,7, Total gates=17 //
procedure InvRND05(var _in, _out);
var
  t02, t03, t04, t05, 
  t07, t08, t09, t10, 
  t12, t13, t15, t16, 
  t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] and Bi[3];
  t02 := Bi[2] xor t01;
  t03 := Bi[0] xor Bi[3];
  t04 := Bi[1] and t02;
  t05 := Bi[0] and Bi[2];
  Bo[0] := t03 xor t04;
  t07 := Bi[0] and Bo[0];
  t08 := t01   xor Bo[0];
  t09 := Bi[1] or t05;
  t10 :=       not Bi[1];
  Bo[1] := t08 xor t09;
  t12 := t10   or t07;
  t13 := Bo[0] or Bo[1];
  Bo[3] := t02 xor t12;
  t15 := t02   xor t13;
  t16 := Bi[1] xor Bi[3];
  Bo[2] := t16 xor t15;
end;

// S6:   7  2 12  5  8  4  6 11 14  9  1 15 13  3 10  0 //
// depth = 8,3,6,3, Total gates=19 //
procedure RND06(var _in, _out);
var
  t02, t03, t04, t05, 
  t07, t08, t09, t10, 
  t11, t12, t13, t15, 
  t17, t18, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] and Bi[3];
  t02 := Bi[1] xor Bi[2];
  t03 := Bi[0] xor Bi[3];
  t04 := t01   xor t02;
  t05 := Bi[1] or Bi[2];
  Bo[1] :=     not t04;
  t07 := t03   and t05;
  t08 := Bi[1] and Bo[1];
  t09 := Bi[0] or Bi[2];
  t10 := t07   xor t08;
  t11 := Bi[1] or Bi[3];
  t12 := Bi[2] xor t11;
  t13 := t09   xor t10;
  Bo[2] :=     not t13;
  t15 := Bo[1] and t03;
  Bo[3] := t12 xor t07;
  t17 := Bi[0] xor Bi[1];
  t18 := Bo[2] xor t15;
  Bo[0] := t17 xor t18;
end;


// InvS6:  15 10  1 13  5  3  6  0  4  9 14  7  2 12  8 11 //
// depth = 5,3,8,6, Total gates=19 //
procedure InvRND06(var _in, _out);
var
  t02, t03, t04, t05, 
  t06, t07, t08, t09, 
  t12, t13, t14, t15, 
  t16, t17, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] xor Bi[2];
  t02 :=       not Bi[2];
  t03 := Bi[1] and t01;
  t04 := Bi[1] or t02;
  t05 := Bi[3] or t03;
  t06 := Bi[1] xor Bi[3];
  t07 := Bi[0] and t04;
  t08 := Bi[0] or t02;
  t09 := t07   xor t05;
  Bo[1] := t06 xor t08;
  Bo[0] :=     not t09;
  t12 := Bi[1] and Bo[0];
  t13 := t01   and t05;
  t14 := t01   xor t12;
  t15 := t07   xor t13;
  t16 := Bi[3] or t02;
  t17 := Bi[0] xor Bo[1];
  Bo[3] := t17 xor t15;
  Bo[2] := t16 xor t14;
end;

// S7:   1 13 15  0 14  8  2 11  7  4 12 10  9  3  5  6 //
// depth = 10,7,10,4, Total gates=19 //
procedure RND07(var _in, _out);
var
  t02, t03, t04, t05, 
  t06, t08, t09, t10, 
  t11, t13, t14, t15, 
  t16, t17, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] and Bi[2];
  t02 :=       not Bi[3];
  t03 := Bi[0] and t02;
  t04 := Bi[1] or t01;
  t05 := Bi[0] and Bi[1];
  t06 := Bi[2] xor t04;
  Bo[3] := t03 xor t06;
  t08 := Bi[2] or Bo[3];
  t09 := Bi[3] or t05;
  t10 := Bi[0] xor t08;
  t11 := t04   and Bo[3];
  Bo[1] := t09 xor t10;
  t13 := Bi[1] xor Bo[1];
  t14 := t01   xor Bo[1];
  t15 := Bi[2] xor t05;
  t16 := t11   or t13;
  t17 := t02   or t14;
  Bo[0] := t15 xor t17;
  Bo[2] := Bi[0] xor t16;
end;

// InvS7:   3  0  6 13  9 14 15  8  5 12 11  7 10  1  4  2 //
// depth = 9,7,3,3, Total gates=18 //
procedure InvRND07(var _in, _out);
var
  t02, t03, t04, t06, 
  t07, t08, t09, t10, 
  t11, t13, t14, t15, 
  t16, t01: LongWord;
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  t01 := Bi[0] and Bi[1];
  t02 := Bi[0] or Bi[1];
  t03 := Bi[2] or t01;
  t04 := Bi[3] and t02;
  Bo[3] := t03 xor t04;
  t06 := Bi[1] xor t04;
  t07 := Bi[3] xor Bo[3];
  t08 :=       not t07;
  t09 := t06   or t08;
  t10 := Bi[1] xor Bi[3];
  t11 := Bi[0] or Bi[3];
  Bo[1] := Bi[0] xor t09;
  t13 := Bi[2] xor t06;
  t14 := Bi[2] and t11;
  t15 := Bi[3] or Bo[1];
  t16 := t01   or t10;
  Bo[0] := t13 xor t15;
  Bo[2] := t14 xor t16;
end;

procedure Transform(var _in, _out);
var
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  Bo[0] := (Bi[0] shl 13) or (Bi[0] shr 19); //ROL(Bi[0], 13);
  Bo[2] := (Bi[2] shl 3) or (Bi[2] shr 29); //ROL(Bi[2], 3);
  Bo[1] := Bi[1] xor Bo[0] xor Bo[2];
  Bo[3] := Bi[3] xor Bo[2] xor Bo[0] shl 3;
  Bo[1] := (Bo[1] shl 1) or (Bo[1] shr 31); //ROL(Bo[1], 1);
  Bo[3] := (Bo[3] shl 7) or (Bo[3] shr 25); //ROL(Bo[3], 7);
  Bo[0] := Bo[0] xor Bo[1] xor Bo[3];
  Bo[2] := Bo[2] xor Bo[3] xor (Bo[1] shl 7);
  Bo[0] := (Bo[0] shl 5) or (Bo[0] shr 27); //ROL(Bo[0], 5);
  Bo[2] := (Bo[2] shl 22) or (Bo[2] shr 10); //ROL(Bo[2], 22);
end;

procedure InvTransform(var _in, _out);
var
  Bi: TSerpentBlock absolute _in;
  Bo: TSerpentBlock absolute _out;
begin
  Bo[2] := (Bi[2] shr 22) or (Bi[2] shl 10); //ROR(Bi[2], 22);
  Bo[0] := (Bi[0] shr 5) or (Bi[0] shl 27); //ROR(Bi[0], 5);
  Bo[2] := Bo[2] xor Bi[3] xor (Bi[1] shl 7);
  Bo[0] := Bo[0] xor Bi[1] xor Bi[3];
  Bo[3] := (Bi[3] shr 7) or (Bi[3] shl 25); //ROR(Bi[3], 7);
  Bo[1] := (Bi[1] shr 1) or (Bi[1] shl 31); //ROR(Bi[1], 1);
  Bo[3] := Bo[3] xor Bo[2] xor Bo[0] shl 3;
  Bo[1] := Bo[1] xor Bo[0] xor Bo[2];
  Bo[2] := (Bo[2] shr 3) or (Bo[2] shl 29); //ROR(Bo[2], 3);
  Bo[0] := (Bo[0] shr 13) or (Bo[0] shl 19); //ROR(Bo[0], 13);
end;


constructor ThiSerpentCipher.Create;
begin
  FBlockSize := BLOCK_SIZE;
  inherited;
end;


procedure ThiSerpentCipher.InitializeKey(Key: Pointer; KeyLen: Integer);
var
  i: Integer;
  w: array [0..SUBKEY_COUNT-1] of LongWord;
begin

  if KeyLen > MAX_KEY_LENGTH then KeyLen := MAX_KEY_LENGTH;
  FillChar(w[0], MAX_KEY_LENGTH, #0);
  Move(Key^, w[0], KeyLen);

  
  //    32  -    ,
  //  = 0
  if KeyLen < MAX_KEY_LENGTH then
  begin
    i := KeyLen div 4; //   w[],   
    w[i] := w[i] or (1 shl ((KeyLen mod 4) * 8)); // (KeyLen mod 4) * 8 =    w[i]   
  end;

  
  for i := 8 to 15 do
    w[i] := ROL(w[i-8] xor w[i-5] xor w[i-3] xor w[i-1] xor PHI xor LongWord(i-8), 11);
  
  for i := 0 to 7 do
    w[i] := w[i+8];
    
  for i := 8 to SUBKEY_COUNT-1 do
    w[i] := ROL(w[i-8] xor w[i-5] xor w[i-3] xor w[i-1] xor PHI xor LongWord(i), 11);

  
  
  RND03(w[  0], SubKeys[  0]);
  RND02(w[  4], SubKeys[  4]);
  RND01(w[  8], SubKeys[  8]);
  RND00(w[ 12], SubKeys[ 12]);
  RND07(w[ 16], SubKeys[ 16]);
  RND06(w[ 20], SubKeys[ 20]);
  RND05(w[ 24], SubKeys[ 24]);
  RND04(w[ 28], SubKeys[ 28]);
  RND03(w[ 32], SubKeys[ 32]);
  RND02(w[ 36], SubKeys[ 36]);
  RND01(w[ 40], SubKeys[ 40]);
  RND00(w[ 44], SubKeys[ 44]);
  RND07(w[ 48], SubKeys[ 48]);
  RND06(w[ 52], SubKeys[ 52]);
  RND05(w[ 56], SubKeys[ 56]);
  RND04(w[ 60], SubKeys[ 60]);
  RND03(w[ 64], SubKeys[ 64]);
  RND02(w[ 68], SubKeys[ 68]);
  RND01(w[ 72], SubKeys[ 72]);
  RND00(w[ 76], SubKeys[ 76]);
  RND07(w[ 80], SubKeys[ 80]);
  RND06(w[ 84], SubKeys[ 84]);
  RND05(w[ 88], SubKeys[ 88]);
  RND04(w[ 92], SubKeys[ 92]);
  RND03(w[ 96], SubKeys[ 96]);
  RND02(w[100], SubKeys[100]);
  RND01(w[104], SubKeys[104]);
  RND00(w[108], SubKeys[108]);
  RND07(w[112], SubKeys[112]);
  RND06(w[116], SubKeys[116]);
  RND05(w[120], SubKeys[120]);
  RND04(w[124], SubKeys[124]);
  RND03(w[128], SubKeys[128]);

end;

type
  TRoundProc = procedure (var _in, _out);
  
const
  RoundProc: array [0..ROUNDS-2] of TRoundProc =
    (
      RND00, RND01, RND02, RND03, 
      RND04, RND05, RND06, RND07, 
      RND00, RND01, RND02, RND03, 
      RND04, RND05, RND06, RND07, 
      RND00, RND01, RND02, RND03, 
      RND04, RND05, RND06, RND07, 
      RND00, RND01, RND02, RND03, 
      RND04, RND05, RND06 
    );


procedure ThiSerpentCipher.EncryptBlock(var Block);
var
  Bi: TSerpentBlock absolute Block;
  Bo: TSerpentBlock;
  i: Integer;
begin
  for i := 0 to ROUNDS-2 do
  begin
    Bi[0] := Bi[0] xor SubKeys[i*4    ];
    Bi[1] := Bi[1] xor SubKeys[i*4 + 1];
    Bi[2] := Bi[2] xor SubKeys[i*4 + 2];
    Bi[3] := Bi[3] xor SubKeys[i*4 + 3];
    RoundProc[i](Bi, Bo);
    Transform(Bo, Bi);
  end;
  // Round 31
  Bi[0] := Bi[0] xor SubKeys[(ROUNDS-1)*4 + 0];
  Bi[1] := Bi[1] xor SubKeys[(ROUNDS-1)*4 + 1];
  Bi[2] := Bi[2] xor SubKeys[(ROUNDS-1)*4 + 2];
  Bi[3] := Bi[3] xor SubKeys[(ROUNDS-1)*4 + 3];
  RND07(Bi, Bo);
  
  Bi[0] := Bo[0] xor SubKeys[ROUNDS * 4 + 0];
  Bi[1] := Bo[1] xor SubKeys[ROUNDS * 4 + 1];
  Bi[2] := Bo[2] xor SubKeys[ROUNDS * 4 + 2];
  Bi[3] := Bo[3] xor SubKeys[ROUNDS * 4 + 3];
end;



const
  InvRoundProc: array [0..ROUNDS-1] of TRoundProc =
    (
      nil,      InvRND01, InvRND02, InvRND03, 
      InvRND04, InvRND05, InvRND06, InvRND07,
      InvRND00, InvRND01, InvRND02, InvRND03, 
      InvRND04, InvRND05, InvRND06, InvRND07,
      InvRND00, InvRND01, InvRND02, InvRND03, 
      InvRND04, InvRND05, InvRND06, InvRND07,
      InvRND00, InvRND01, InvRND02, InvRND03, 
      InvRND04, InvRND05, InvRND06, InvRND07
    );

procedure ThiSerpentCipher.DecryptBlock(var Block);
var
  Bi: TSerpentBlock absolute Block;
  Bo: TSerpentBlock;
  i: Integer;
begin
  Bi[0] := Bi[0] xor SubKeys[ROUNDS*4    ];
  Bi[1] := Bi[1] xor SubKeys[ROUNDS*4 + 1];
  Bi[2] := Bi[2] xor SubKeys[ROUNDS*4 + 2];
  Bi[3] := Bi[3] xor SubKeys[ROUNDS*4 + 3];
  
  for i := ROUNDS-1 downto 1 do
  begin
    InvRoundProc[i](Bi, Bo);
    Bo[0] := Bo[0] xor SubKeys[i*4    ];
    Bo[1] := Bo[1] xor SubKeys[i*4 + 1];
    Bo[2] := Bo[2] xor SubKeys[i*4 + 2];
    Bo[3] := Bo[3] xor SubKeys[i*4 + 3];
    InvTransform(Bo, Bi);
  end;
  
  // Round 0
  InvRND00(Bi, Bo);
  Bi[0] := Bo[0] xor SubKeys[0];
  Bi[1] := Bo[1] xor SubKeys[1];
  Bi[2] := Bo[2] xor SubKeys[2];
  Bi[3] := Bo[3] xor SubKeys[3];
end;

end.