// ========================================================================================= //
//                                                                                           //
//                           Twofish                          //
//                                     2016-01-18                                      //
//                                                                                           //
//                                       HiAsm                                         //
//                                     www.hiasm.com                                         //
//                                                                                           //
//        : Netspirit                                                                   //
//        E-mail: netspirit@meta.ua                                                          //
//                                                                                           //
//                                                       //
//                                                                                           //
//        1)    " ".                      //
//              .                                                //
//        2)          ,                     //
//                ,                  //
//            .                                                                     //
//        3)                             //
//           (  )   .                            //
//                                //
//                .                                //
//        4)                          //
//                                   //
//                  .                   //
//        5)        (, )         //
//                .                                //
//                                                                                           //
//                        ,                              //
//                                                                  //
//                                                                                           //
//                                           //
//                                                                                           //
// ========================================================================================= //

                  // ************************************************* //
                  //               TwoFish Algorithm                   //
                  //  Copyright 1998, Hi/fn and Counterpane Systems.   //
                  // ************************************************* //



unit hiTwofishCipher;



interface

uses Windows, Kol, Share, BlockCipher;

const
  ROUNDS = 16;
  
  BLOCK_SIZE = 16; // bytes in a data-block
  
  // Subkey array indices
  INPUT_WHITEN = 0;
  OUTPUT_WHITEN = INPUT_WHITEN +  BLOCK_SIZE div 4;
  ROUND_SUBKEYS = OUTPUT_WHITEN + BLOCK_SIZE div 4; // 2*(# rounds)
  
  SUBKEY_COUNT = ROUND_SUBKEYS + 2*ROUNDS;


type

  TTwofishBlock = array [0..(BLOCK_SIZE div 4)-1] of Integer;
  
  ThiTwofishCipher = class(ThiBlockCipher)
    private
      SubKeys: array of Integer;
      SBox: array [0..4*256-1] of Integer;
      
      function Fe32(x: Integer): Integer;
      
    public
      constructor Create;
      procedure EncryptBlock(var Block); override;
      procedure DecryptBlock(var Block); override;
      procedure InitializeKey(Key: Pointer; KeyLen: Integer); override;
  end;


implementation



const

  // Fixed 8x8 permutation S-boxes
  P: array [0..1] of array [0..255] of Byte = (
    (  // p0
       $A9, $67, $B3, $E8, $04, $FD, $A3, $76,
       $9A, $92, $80, $78, $E4, $DD, $D1, $38,
       $0D, $C6, $35, $98, $18, $F7, $EC, $6C,
       $43, $75, $37, $26, $FA, $13, $94, $48,
       $F2, $D0, $8B, $30, $84, $54, $DF, $23,
       $19, $5B, $3D, $59, $F3, $AE, $A2, $82,
       $63, $01, $83, $2E, $D9, $51, $9B, $7C,
       $A6, $EB, $A5, $BE, $16, $0C, $E3, $61,
       $C0, $8C, $3A, $F5, $73, $2C, $25, $0B,
       $BB, $4E, $89, $6B, $53, $6A, $B4, $F1,
       $E1, $E6, $BD, $45, $E2, $F4, $B6, $66,
       $CC, $95, $03, $56, $D4, $1C, $1E, $D7,
       $FB, $C3, $8E, $B5, $E9, $CF, $BF, $BA,
       $EA, $77, $39, $AF, $33, $C9, $62, $71,
       $81, $79, $09, $AD, $24, $CD, $F9, $D8,
       $E5, $C5, $B9, $4D, $44, $08, $86, $E7,
       $A1, $1D, $AA, $ED, $06, $70, $B2, $D2,
       $41, $7B, $A0, $11, $31, $C2, $27, $90,
       $20, $F6, $60, $FF, $96, $5C, $B1, $AB,
       $9E, $9C, $52, $1B, $5F, $93, $0A, $EF,
       $91, $85, $49, $EE, $2D, $4F, $8F, $3B,
       $47, $87, $6D, $46, $D6, $3E, $69, $64,
       $2A, $CE, $CB, $2F, $FC, $97, $05, $7A,
       $AC, $7F, $D5, $1A, $4B, $0E, $A7, $5A,
       $28, $14, $3F, $29, $88, $3C, $4C, $02,
       $B8, $DA, $B0, $17, $55, $1F, $8A, $7D,
       $57, $C7, $8D, $74, $B7, $C4, $9F, $72,
       $7E, $15, $22, $12, $58, $07, $99, $34,
       $6E, $50, $DE, $68, $65, $BC, $DB, $F8,
       $C8, $A8, $2B, $40, $DC, $FE, $32, $A4,
       $CA, $10, $21, $F0, $D3, $5D, $0F, $00,
       $6F, $9D, $36, $42, $4A, $5E, $C1, $E0
    ),
    (  // p1
       $75, $F3, $C6, $F4, $DB, $7B, $FB, $C8,
       $4A, $D3, $E6, $6B, $45, $7D, $E8, $4B,
       $D6, $32, $D8, $FD, $37, $71, $F1, $E1,
       $30, $0F, $F8, $1B, $87, $FA, $06, $3F,
       $5E, $BA, $AE, $5B, $8A, $00, $BC, $9D,
       $6D, $C1, $B1, $0E, $80, $5D, $D2, $D5,
       $A0, $84, $07, $14, $B5, $90, $2C, $A3,
       $B2, $73, $4C, $54, $92, $74, $36, $51,
       $38, $B0, $BD, $5A, $FC, $60, $62, $96,
       $6C, $42, $F7, $10, $7C, $28, $27, $8C,
       $13, $95, $9C, $C7, $24, $46, $3B, $70,
       $CA, $E3, $85, $CB, $11, $D0, $93, $B8,
       $A6, $83, $20, $FF, $9F, $77, $C3, $CC,
       $03, $6F, $08, $BF, $40, $E7, $2B, $E2,
       $79, $0C, $AA, $82, $41, $3A, $EA, $B9,
       $E4, $9A, $A4, $97, $7E, $DA, $7A, $17,
       $66, $94, $A1, $1D, $3D, $F0, $DE, $B3,
       $0B, $72, $A7, $1C, $EF, $D1, $53, $3E,
       $8F, $33, $26, $5F, $EC, $76, $2A, $49,
       $81, $88, $EE, $21, $C4, $1A, $EB, $D9,
       $C5, $39, $99, $CD, $AD, $31, $8B, $01,
       $18, $23, $DD, $1F, $4E, $2D, $F9, $48,
       $4F, $F2, $65, $8E, $78, $5C, $58, $19,
       $8D, $E5, $98, $57, $67, $7F, $05, $64,
       $AF, $63, $B6, $FE, $F5, $B7, $3C, $A5,
       $CE, $E9, $68, $44, $E0, $4D, $43, $69,
       $29, $2E, $AC, $15, $59, $A8, $0A, $9E,
       $6E, $47, $DF, $34, $35, $6A, $CF, $DC,
       $22, $C9, $C0, $9B, $89, $D4, $ED, $AB,
       $12, $A2, $0D, $52, $BB, $02, $2F, $A9,
       $D7, $61, $1E, $B4, $50, $04, $F6, $C2,
       $16, $25, $86, $56, $55, $09, $BE, $91
    )
  );


    {
      * Define the fixed p0/p1 permutations used in keyed S-box lookup.
      * By changing the following constant definitions, the S-boxes will
      * automatically get changed in the Twofish engine.
    }
   P_00 = 1;
   P_01 = 0;
   P_02 = 0;
   P_03 = P_01 xor 1;
   P_04 = 1;

   P_10 = 0;
   P_11 = 0;
   P_12 = 1;
   P_13 = P_11 xor 1;
   P_14 = 0;

   P_20 = 1;
   P_21 = 1;
   P_22 = 0;
   P_23 = P_21 xor 1;
   P_24 = 0;

   P_30 = 0;
   P_31 = 1;
   P_32 = 1;
   P_33 = P_31 xor 1;
   P_34 = 1;
   
   
  SK_STEP = $02020202;
  SK_BUMP = $01010101;
  SK_ROTL = 9;
   
   
   // Primitive polynomial for GF(256) 
   GF256_FDBK =   $169;
   GF256_FDBK_2 = $169 div 2;
   GF256_FDBK_4 = $169 div 4;
   
   RS_GF_FDBK = $14D; // field generator
   
   MAX_KEY_LENGTH = 32; //     
   MAX_KEY64_IN_KEY = MAX_KEY_LENGTH div 8; //  8-   


var
  // MDS matrix
  MDS: array[0..3] of array[0..255] of Integer;

  
function LFSR1(x: Integer): Integer;
begin
  Result := (x shr 1);
  if (x and $01) <> 0 then Result := Result xor GF256_FDBK_2;
end;

function LFSR2(x: Integer): Integer;
begin
  Result := x shr 2;
  if (x and $02) <> 0 then Result := Result xor GF256_FDBK_2;
  if (x and $01) <> 0 then Result := Result xor GF256_FDBK_4;
end;

function Mx_X(x: Integer): Integer;
begin
  Result := x xor LFSR2(x); // 5B
end;

function Mx_Y(x: Integer): Integer;
begin
  Result := x xor LFSR1(x) xor LFSR2(x); // EF
end;

procedure InitializeMDS;
var
  m1, mX, mY: array [0..1] of Integer;
  i, j: Integer;
begin

  for i := 0 to 255 do
  begin
    j := P[0][i]; // Compute all the matrix elements
    m1[0] := j;
    mX[0] := Mx_X(j);
    mY[0] := Mx_Y(j);
    
    j := P[1][i];
    m1[1] := j;
    mX[1] := Mx_X(j);
    mY[1] := Mx_Y(j);
    
     MDS[0][i] := (m1[P_00] {shl  0}) or // Fill matrix w/ above elements
                  (mX[P_00] shl  8) or
                  (mY[P_00] shl 16) or
                  (mY[P_00] shl 24);
     MDS[1][i] := (mY[P_10] {shl  0}) or
                  (mY[P_10] shl  8) or
                  (mX[P_10] shl 16) or
                  (m1[P_10] shl 24);
     MDS[2][i] := (mX[P_20] {shl  0}) or
                  (mY[P_20] shl  8) or
                  (m1[P_20] shl 16) or
                  (mY[P_20] shl 24);
     MDS[3][i] := (mX[P_30] {shl  0}) or
                  (m1[P_30] shl  8) or
                  (mY[P_30] shl 16) or
                  (mX[P_30] shl 24);
  end;

end;


{
  * Reed-Solomon code parameters: (12, 8) reversible code:<p>
  * <pre>
  *   g(x) = x**4 + (a + 1/a) x**3 + a x**2 + (a + 1/a) x + 1
  * </pre>
  * where a = primitive root of field generator 0x14D
}

function RS_rem(x: Integer): Integer;
var
  b, g2, g3: Integer;
begin
  b := (x shr 24) and $FF;
  g2 := b shl 1;
  if (b and $80) <> 0 then g2 := g2 xor RS_GF_FDBK;
  g2 := g2 and $FF;
  
  g3 := b shr 1;
  if (b and $01) <> 0 then g3 := g3 xor (RS_GF_FDBK shr 1);
  g3 := g3 xor g2;
  
  Result := (x shl 8) xor (g3 shl 24) xor (g2 shl 16) xor (g3 shl 8) xor b;
end;

{
  * Use (12, 8) Reed-Solomon code over GF(256) to produce a key S-box
  * 32-bit entity from two key material 32-bit entities.
  *
  * @param  k0  1st 32-bit entity.
  * @param  k1  2nd 32-bit entity.
  * @return  Remainder polynomial generated using RS code
}

function RS_MDS_Encode(const k0, k1: Integer): Integer;
var
  i: Integer;
begin
  Result := k1;
  for i := 0 to 3 do
    Result := RS_rem(Result);
  Result := Result xor k0;
  for i := 0 to 3 do
    Result := RS_rem(Result);
end;



  
function F32(k64Cnt, x: Integer; k32: array of Integer): Integer;
var
  b0, b1, b2, b3: Integer;
  k0, k1, k2, k3: Integer;
begin
  b0 := x and $FF;
  b1 := x shr 8 and $FF;
  b2 := x shr 16 and $FF;
  b3 := x shr 24 and $FF;
  
  k0 := k32[0];
  k1 := k32[1];
  k2 := k32[2];
  k3 := k32[3];
  

  if  k64Cnt = 1 then // 64 bit key
  begin
      Result := 
            MDS[0][(P[P_01][b0]) xor (k0 and $FF)] xor
            MDS[1][(P[P_11][b1]) xor (k0 shr 8 and $FF)] xor
            MDS[2][(P[P_21][b2]) xor (k0 shr 16 and $FF)] xor
            MDS[3][(P[P_31][b3]) xor (k0 shr 24 and $FF)];
  end
  else // 128, 192, 256 bit key
  begin
    
    if  k64Cnt = 4 then // 256 bit key
    begin
      b0 := (P[P_04][b0]) xor (k3 and $FF);
      b1 := (P[P_14][b1]) xor (k3 shr 8 and $FF);
      b2 := (P[P_24][b2]) xor (k3 shr 16 and $FF);
      b3 := (P[P_34][b3]) xor (k3 shr 24 and $FF);
    end;

    if (k64Cnt = 4) or (k64Cnt = 3) then // 128, 256 bit key
    begin
      b0 := (P[P_03][b0]) xor (k2 and $FF);
      b1 := (P[P_13][b1]) xor (k2 shr 8 and $FF);
      b2 := (P[P_23][b2]) xor (k2 shr 16 and $FF);
      b3 := (P[P_33][b3]) xor (k2 shr 24 and $FF);
    end;
    
    Result := 
            MDS[0][(P[P_01][(P[P_02][b0]) xor (k1 and $FF)]) xor (k0 and $FF)] xor
            MDS[1][(P[P_11][(P[P_12][b1]) xor (k1 shr 8 and $FF)]) xor (k0 shr 8 and $FF)] xor
            MDS[2][(P[P_21][(P[P_22][b2]) xor (k1 shr 16 and $FF)]) xor (k0 shr 16 and $FF)] xor
            MDS[3][(P[P_31][(P[P_32][b3]) xor (k1 shr 24 and $FF)]) xor (k0 shr 24 and $FF)];
  end;
end;

constructor ThiTwofishCipher.Create;
begin
  FBlockSize := BLOCK_SIZE;
  inherited;
end;

procedure ThiTwofishCipher.InitializeKey(Key: Pointer; KeyLen: Integer);
var
  k64Cnt: Integer;
  i, j: Integer;
  A, B: Integer;
  
  b0, b1, b2, b3: Integer;
  k0, k1, k2, k3: Integer;
  
  key32: array of Integer;
  k32e: array [0..MAX_KEY64_IN_KEY-1] of Integer; // even 32-bit entities
  k32o: array [0..MAX_KEY64_IN_KEY-1] of Integer; // odd 32-bit entities
  sBoxKey: array [0..MAX_KEY64_IN_KEY-1] of Integer;

begin
  
  if KeyLen > MAX_KEY_LENGTH then KeyLen := MAX_KEY_LENGTH;
  k64Cnt := KeyLen div 8;
  if (k64Cnt * 8 < KeyLen) or (KeyLen = 0) then Inc(k64Cnt);

  SetLength(key32, k64Cnt * 2);
  FillChar(key32[0], k64Cnt * 8, #0); // 8-    
  //key32[High(key32)] := 0;
  //key32[High(key32)-1] := 0;
  
  Move(Key^, key32[0], KeyLen);


  // Split user key material into even and odd 32-bit entities and
  // compute S-box keys using (12, 8) Reed-Solomon code over GF(256)

  j := k64Cnt-1;
  for i := 0 to k64Cnt-1 do
  begin
    k32e[i] := key32[2*i];
    k32o[i] := key32[2*i+1];
    
    sBoxKey[j] := RS_MDS_Encode(k32e[i], k32o[i]); // reverse order

    Dec(j);
  end;
  
  // Compute the round decryption subkeys for PHT. These same subkeys
  // will be used in encryption but will be applied in reverse order.
  SetLength(SubKeys, SUBKEY_COUNT);
  j := 0;
  for i := 0 to (SUBKEY_COUNT div 2)-1 do
  begin
    A := F32(k64Cnt, j        , k32e); // A uses even key entities
    B := F32(k64Cnt, j+SK_BUMP, k32o); // B uses odd  key entities
    B := (B shl 8) or (B shr 24);
    Inc(A, B);
    SubKeys[2*i] := A;               // combine with a PHT
    Inc(A, B);
    SubKeys[2*i + 1] := (A shl SK_ROTL) or (A shr (32-SK_ROTL));
    Inc(j, SK_STEP);
  end;
  

  // Fully expand the table for speed
  
  k0 := sBoxKey[0];
  k1 := sBoxKey[1];
  k2 := sBoxKey[2];
  k3 := sBoxKey[3];
  
  for i := 0 to 255 do
  begin
    b0 := i;
    b1 := i;
    b2 := i;
    b3 := i;

    if k64Cnt = 1 then
    begin
      SBox[     2*i  ] := MDS[0][(P[P_01][b0]) xor (k0 and $FF)];
      SBox[     2*i+1] := MDS[1][(P[P_11][b1]) xor (k0 shr 8 and $FF)];
      SBox[$200+2*i  ] := MDS[2][(P[P_21][b2]) xor (k0 shr 16 and $FF)];
      SBox[$200+2*i+1] := MDS[3][(P[P_31][b3]) xor (k0 shr 24 and $FF)];
    end
    else
    begin
      if k64Cnt = 4 then 
      begin
        b0 := (P[P_04][b0]) xor (k3 and $FF);
        b1 := (P[P_14][b1]) xor (k3 shr 8 and $FF);
        b2 := (P[P_24][b2]) xor (k3 shr 16 and $FF);
        b3 := (P[P_34][b3]) xor (k3 shr 24 and $FF);
      end;
      if (k64Cnt = 4) or (k64Cnt = 3) then 
      begin
        b0 := (P[P_03][b0]) xor (k2 and $FF);
        b1 := (P[P_13][b1]) xor (k2 shr 8 and $FF);
        b2 := (P[P_23][b2]) xor (k2 shr 16 and $FF);
        b3 := (P[P_33][b3]) xor (k2 shr 24 and $FF);
      end;
      
      SBox[     2*i  ] := MDS[0][(P[P_01][(P[P_02][b0]) xor (k1 and $FF)]) xor (k0 and $FF)];
      SBox[     2*i+1] := MDS[1][(P[P_11][(P[P_12][b1]) xor (k1 shr 8 and $FF)]) xor (k0 shr 8 and $FF)];
      SBox[$200+2*i  ] := MDS[2][(P[P_21][(P[P_22][b2]) xor (k1 shr 16 and $FF)]) xor (k0 shr 16 and $FF)];
      SBox[$200+2*i+1] := MDS[3][(P[P_31][(P[P_32][b3]) xor (k1 shr 24 and $FF)]) xor (k0 shr 24 and $FF)];
    end;
    
  end;
end;
  
function ThiTwofishCipher.Fe32(x: Integer): Integer;
begin
  Result := SBox[       2 * (x and $FF)    ] xor
            SBox[       2 * (x shr 8 and $FF) + 1] xor
            SBox[$200 + 2 * (x shr 16 and $FF)    ] xor
            SBox[$200 + 2 * (x shr 24 and $FF) + 1];
end;

procedure ThiTwofishCipher.EncryptBlock(var Block);
var
  B: TTwofishBlock absolute Block;
  x0, x1, x2, x3: Integer;
  
  t0, t1, k: Integer;
  R: Integer;
begin

  x0 := B[0] xor SubKeys[INPUT_WHITEN];
  x1 := B[1] xor SubKeys[INPUT_WHITEN + 1];
  x2 := B[2] xor SubKeys[INPUT_WHITEN + 2];
  x3 := B[3] xor SubKeys[INPUT_WHITEN + 3];
  
  k := ROUND_SUBKEYS;
  
  R := 0;
  while R < ROUNDS do
  begin
    t0 := Fe32(x0);
    t1 := Fe32((x1 shl 8) or (x1 shr 24)); // ROL(x1, 8);
    x2 := x2 xor (t0 + t1 + SubKeys[k]);
    Inc(k);
    x2 := (x2 shr 1) or (x2 shl 31); //x2 := ROR(x2, 1);
    x3 := (x3 shl 1) or (x3 shr 31); //x3 := ROL(x3, 1);
    x3 := x3 xor (t0 + 2*t1 + SubKeys[k]);
    Inc(k);
    
    t0 := Fe32(x2);
    t1 := Fe32((x3 shl 8) or (x3 shr 24)); // ROL(x1, 8);
    x0 := x0 xor (t0 + t1 + SubKeys[k]);
    Inc(k);
    x0 := (x0 shr 1) or (x0 shl 31); //x0 := ROR(x0, 1);
    x1 := (x1 shl 1) or (x1 shr 31); //x1 := ROL(x1, 1);
    x1 := x1 xor (t0 + 2*t1 + SubKeys[k]);
    Inc(k);
    
    Inc(R, 2);
  end;
  

  B[0] := x2 xor SubKeys[OUTPUT_WHITEN];
  B[1] := x3 xor SubKeys[OUTPUT_WHITEN + 1];
  B[2] := x0 xor SubKeys[OUTPUT_WHITEN + 2];
  B[3] := x1 xor SubKeys[OUTPUT_WHITEN + 3]; 
end;

procedure ThiTwofishCipher.DecryptBlock(var Block);
var
  B: TTwofishBlock absolute Block;
  x0, x1, x2, x3: Integer;
  
  t0, t1, k: Integer;
  R: Integer;
begin

  x2 := B[0] xor SubKeys[OUTPUT_WHITEN];
  x3 := B[1] xor SubKeys[OUTPUT_WHITEN + 1];
  x0 := B[2] xor SubKeys[OUTPUT_WHITEN + 2];
  x1 := B[3] xor SubKeys[OUTPUT_WHITEN + 3];
  
  
  k := SUBKEY_COUNT - 1;
  
  R := 0;
  while R < ROUNDS do
  begin
    t0 := Fe32(x2);
    t1 := Fe32((x3 shl 8) or (x3 shr 24));
    x1 := x1 xor (t0 + 2*t1 + SubKeys[k]);
    Dec(k);
    x1 := (x1 shr 1) or (x1 shl 31);
    x0 := (x0 shl 1) or (x0 shr 31);
    x0 := x0 xor (t0 + t1 + SubKeys[k]);
    Dec(k);
    
    t0 := Fe32(x0);
    t1 := Fe32((x1 shl 8) or (x1 shr 24));
    x3 := x3 xor (t0 + 2*t1 + SubKeys[k]);
    Dec(k);
    x3 := (x3 shr 1) or (x3 shl 31);
    x2 := (x2 shl 1) or (x2 shr 31);
    x2 := x2 xor (t0 + t1 + SubKeys[k]);
    Dec(k);
    
    Inc(R, 2);
  end;
  
  B[0] := x0 xor SubKeys[INPUT_WHITEN];
  B[1] := x1 xor SubKeys[INPUT_WHITEN + 1];
  B[2] := x2 xor SubKeys[INPUT_WHITEN + 2];
  B[3] := x3 xor SubKeys[INPUT_WHITEN + 3];
end;





initialization
  InitializeMDS;

end.